from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.apps.virtualization.vmware_workstation import (
    VmwareWorkstationPlugin,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("target", "fs", "dnd_path"),
    [
        ("target_win_users", "fs_win", "Users\\John\\AppData\\Local\\Temp\\VmwareDND"),
        ("target_unix_users", "fs_unix", "/home/user/.cache/vmware/drag_and_drop"),
    ],
)
def test_vmware_workstation_clipboard_dnd(
    target: Target, fs: VirtualFilesystem, dnd_path: str, request: pytest.FixtureRequest
) -> None:
    """Test if we correctly detect drag and drop artifacts on Windows and Unix targets."""

    target = request.getfixturevalue(target)
    fs = request.getfixturevalue(fs)
    fs.map_dir_from_tar(dnd_path, absolute_path("_data/plugins/apps/virtualization/vmware_workstation/dnd.tar"))
    target.add_plugin(VmwareWorkstationPlugin)
    results = sorted(target.vmware.clipboard(), key=lambda r: r.path)
    dnd_path = dnd_path.replace("\\", "/")

    assert len(results) == 3
    assert [str(r.path).replace("C:\\", "").replace("\\", "/") for r in results] == [
        f"{dnd_path}/8lkBf1/wachter.jpg",
        f"{dnd_path}/iUpQHG/credentials.txt",
        f"{dnd_path}/x1G9fJ/example.txt",
    ]
    assert results[0].ts == datetime(2024, 9, 18, 11, 16, 44, tzinfo=timezone.utc)



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import mock_open, patch

import pytest

from dissect.target.plugins.apps.webserver import iis
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem


def test_iis_iis_format(target_win_tzinfo: Target, fs_win: VirtualFilesystem) -> None:
    config_path = absolute_path("_data/plugins/apps/webserver/iis/iis-applicationHost-iis.config")
    data_dir = absolute_path("_data/plugins/apps/webserver/iis/iis-logs-iis")

    fs_win.map_file("windows/system32/inetsrv/config/applicationHost.config", config_path)
    fs_win.map_dir("Users/John/iis-logs", data_dir)

    target_win_tzinfo.add_plugin(iis.IISLogsPlugin)

    records = list(target_win_tzinfo.iis.logs())

    assert len(records) == 10
    assert {str(r.client_ip) for r in records} == {"127.0.0.1", "::1"}
    assert {str(r.site_name) for r in records} == {"W3SVC1"}
    assert {r.service_status_code for r in records} == {"304", "404", "200"}

    # check if metadata fields are present
    assert {r.format for r in records} == {"IIS"}
    assert {str(r.source) for r in records} == {"sysvol\\Users\\John\\iis-logs\\W3SVC1\\u_in211001.log"}
    assert {r.hostname for r in records} == {target_win_tzinfo.hostname}


@pytest.mark.parametrize(
    "iis_config_path",
    [
        "_data/plugins/apps/webserver/iis/iis-applicationHost-w3c.config",
        "_data/plugins/apps/webserver/iis/iis-applicationHost-w3c-logFile-without-directory-attr.config",
    ],
)
def test_iis_w3c_format(target_win: Target, fs_win: VirtualFilesystem, iis_config_path: str) -> None:
    config_path = absolute_path(iis_config_path)
    data_dir = absolute_path("_data/plugins/apps/webserver/iis/iis-logs-w3c")

    fs_win.map_file("windows/system32/inetsrv/config/applicationHost.config", config_path)
    fs_win.map_dir("Users/John/w3c-logs", data_dir)

    target_win.add_plugin(iis.IISLogsPlugin)

    records = list(target_win.iis.logs())

    assert len(records) == 20

    # first 6 records do not have custom fields and server_name is not set
    assert {r.server_name for r in records[:6]} == {None}
    assert not any(hasattr(r, "custom_field_1") for r in records[:6])
    assert not any(hasattr(r, "custom_field_2") for r in records[:6])

    # other records have the custom fields and server_name set
    assert {r.server_name for r in records[6:]} == {"DESKTOP-PJOQLJS"}
    assert all(hasattr(r, "custom_field_1") for r in records[6:])
    assert all(hasattr(r, "custom_field_2") for r in records[6:])

    assert {str(r.client_ip) for r in records} == {"127.0.0.1", "::1"}
    assert {r.site_name for r in records} == {None, "W3SVC1"}
    assert {r.service_status_code for r in records} == {"304", "404"}

    # check if fields with normalised names are present
    assert {str(r.cs_user_agent) for r in records} == {
        (
            "Mozilla/5.0+(Windows+NT+10.0;+Win64;+x64)+AppleWebKit/537.36+(KHTML,+like+Gecko)"
            "+Chrome/93.0.4577.82+Safari/537.36+Edg/93.0.961.52"
        )
    }

    # check if metadata fields are present
    assert {r.format for r in records} == {"W3C"}
    assert {str(r.source) for r in records} == {"C:\\Users\\John\\w3c-logs\\W3SVC1\\u_ex211001_x.log"}
    assert {r.hostname for r in records} == {target_win.hostname}


@pytest.mark.parametrize(
    ("stream", "method"),
    [
        (b"-, -, 10/1/2021, 00:00:00, -, \xa7, ::1, 1, 2, 3, 200, 0, GET, /, -,", "parse_iis_format_log"),
        (b"#Date: -\n#Fields: s-computername\n\xa7", "parse_w3c_format_log"),
    ],
)
def test_iis_nonutf8(target_win_tzinfo: Target, stream: bytes, method: str) -> None:
    # should not crash on invalid bytes like \xa7
    with patch("pathlib.Path.open", new_callable=mock_open, read_data=stream):
        assert next(iter(getattr(iis, method)(target_win_tzinfo, Path("/iis")))).server_name == "\\xa7"


def test_iis_access_iis_format(target_win_tzinfo: Target, fs_win: VirtualFilesystem) -> None:
    tz = timezone(timedelta(hours=-5))
    config_path = absolute_path("_data/plugins/apps/webserver/iis/iis-applicationHost-iis.config")
    data_dir = absolute_path("_data/plugins/apps/webserver/iis/iis-logs-iis")

    fs_win.map_file("windows/system32/inetsrv/config/applicationHost.config", config_path)
    fs_win.map_dir("Users/John/iis-logs", data_dir)

    target_win_tzinfo.add_plugin(iis.IISLogsPlugin)
    results = list(target_win_tzinfo.iis.access())

    assert len(results) == 10

    record = results[0]
    assert record.ts == datetime(2021, 10, 1, 7, 19, 8, tzinfo=tz)
    assert record.remote_ip == "127.0.0.1"
    assert record.remote_user is None
    assert record.method == "GET"
    assert record.uri == "/"
    assert record.protocol is None
    assert record.status_code == 304
    assert record.bytes_sent == 143
    assert record.referer is None
    assert record.useragent is None
    assert record.source == "sysvol/Users/John/iis-logs/W3SVC1/u_in211001.log"


def test_iis_access_w3c_format(target_win: Target, fs_win: VirtualFilesystem) -> None:
    config_path = absolute_path("_data/plugins/apps/webserver/iis/iis-applicationHost-w3c.config")
    data_dir = absolute_path("_data/plugins/apps/webserver/iis/iis-logs-w3c")

    fs_win.map_file("windows/system32/inetsrv/config/applicationHost.config", config_path)
    fs_win.map_dir("Users/John/w3c-logs", data_dir)

    target_win.add_plugin(iis.IISLogsPlugin)

    results = list(target_win.iis.access())
    assert len(results) == 20

    # W3C format type 1: does not have HTTP version or bytes_sent.
    w3c_record_1 = results[0]
    assert w3c_record_1.ts == datetime(2021, 10, 1, 17, 12, 0, tzinfo=timezone.utc)
    assert w3c_record_1.remote_ip == "127.0.0.1"
    assert w3c_record_1.remote_user is None
    assert w3c_record_1.method == "GET"
    assert w3c_record_1.uri == "/"
    assert w3c_record_1.protocol is None
    assert w3c_record_1.status_code == 304
    assert w3c_record_1.bytes_sent is None
    assert w3c_record_1.referer is None
    assert (
        w3c_record_1.useragent
        == "Mozilla/5.0+(Windows+NT+10.0;+Win64;+x64)+AppleWebKit/537.36+(KHTML,+like+Gecko)+Chrome/93.0.4577.82+Safari/537.36+Edg/93.0.961.52"  # noqa: E501
    )
    assert w3c_record_1.source == "C:/Users/John/w3c-logs/W3SVC1/u_ex211001_x.log"

    # W3C format type 2: contains HTTP version
    w3c_record_2 = results[6]
    assert w3c_record_2.ts == datetime(2021, 10, 1, 17, 34, 48, tzinfo=timezone.utc)
    assert w3c_record_2.remote_ip == "::1"
    assert w3c_record_2.remote_user is None
    assert w3c_record_2.method == "GET"
    assert w3c_record_2.uri == "/"
    assert w3c_record_2.protocol == "HTTP/1.1"
    assert w3c_record_2.status_code == 304
    assert w3c_record_2.bytes_sent == 143
    assert w3c_record_2.referer is None
    assert (
        w3c_record_2.useragent
        == "Mozilla/5.0+(Windows+NT+10.0;+Win64;+x64)+AppleWebKit/537.36+(KHTML,+like+Gecko)+Chrome/93.0.4577.82+Safari/537.36+Edg/93.0.961.52"  # noqa: E501
    )
    assert w3c_record_2.source == "C:/Users/John/w3c-logs/W3SVC1/u_ex211001_x.log"

    # W3C format type 3
    w3c_record_3 = results[11]
    assert w3c_record_3.ts == datetime(2021, 10, 1, 18, 2, 47, tzinfo=timezone.utc)
    assert w3c_record_3.remote_ip == "::1"
    assert w3c_record_3.remote_user is None
    assert w3c_record_3.method == "GET"
    assert w3c_record_3.uri == "/another/path+path2"
    assert w3c_record_3.protocol == "HTTP/1.1"
    assert w3c_record_3.status_code == 404
    assert w3c_record_3.bytes_sent == 5125
    assert w3c_record_3.referer is None
    assert (
        w3c_record_3.useragent
        == "Mozilla/5.0+(Windows+NT+10.0;+Win64;+x64)+AppleWebKit/537.36+(KHTML,+like+Gecko)+Chrome/93.0.4577.82+Safari/537.36+Edg/93.0.961.52"  # noqa: E501
    )
    assert w3c_record_3.source == "C:/Users/John/w3c-logs/W3SVC1/u_ex211001_x.log"

    w3c_record = results[4]
    assert w3c_record.query == "param=value"


@pytest.mark.parametrize(
    "map_dir",
    [
        ("inetpub/logs/LogFiles/W3SVC1"),
        ("inetpub/logs/LogFiles"),
        ("Windows/System32/LogFiles/W3SVC1"),
        ("Windows.old/Windows/System32/LogFiles/W3SVC2"),
        ("Resources/Directory/aaa/LogFiles/Web/W3SVC1"),
    ],
)
@pytest.mark.parametrize(
    "log_format",
    [
        ("iis"),
        ("w3c"),
    ],
)
def test_iis_access_noconfig(
    target_win_tzinfo: Target, fs_win: VirtualFilesystem, map_dir: str, log_format: str
) -> None:
    data_dir = absolute_path(f"_data/plugins/apps/webserver/iis/iis-logs-{log_format}/W3SVC1")
    fs_win.map_dir(map_dir, data_dir)
    target_win_tzinfo.add_plugin(iis.IISLogsPlugin)
    results = list(target_win_tzinfo.iis.access())
    assert len(results) > 0


def test_iis_direct_mode() -> None:
    data_path = absolute_path("_data/plugins/apps/webserver/iis/iis-logs-iis/W3SVC1/u_in211001.log")

    target = Target.open_direct([data_path])
    records = list(target.iis.logs())

    assert len(records) == 10



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO

import dissect.util.ts as ts
from dissect.cstruct import cstruct

from dissect.target.helpers.record import TargetRecordDescriptor

if TYPE_CHECKING:
    from collections.abc import Iterator

DefenderQuarantineRecord = TargetRecordDescriptor(
    "filesystem/windows/defender/quarantine",
    [
        ("datetime", "ts"),
        ("string", "quarantine_id"),
        ("string", "scan_id"),
        ("varint", "threat_id"),
        ("string", "detection_type"),
        ("string", "detection_name"),
    ],
)

DefenderFileQuarantineRecord = TargetRecordDescriptor(
    "filesystem/windows/defender/quarantine/file",
    [
        ("datetime", "ts"),
        ("string", "quarantine_id"),
        ("string", "scan_id"),
        ("varint", "threat_id"),
        ("string", "detection_type"),
        ("string", "detection_name"),
        ("string", "detection_path"),
        ("datetime", "creation_time"),
        ("datetime", "last_write_time"),
        ("datetime", "last_accessed_time"),
        ("string", "resource_id"),
    ],
)

# Source: https://github.com/brad-sp/cuckoo-modified/blob/master/lib/cuckoo/common/quarantine.py#L188
# fmt: off
DEFENDER_QUARANTINE_RC4_KEY = [
    0x1E, 0x87, 0x78, 0x1B, 0x8D, 0xBA, 0xA8, 0x44, 0xCE, 0x69, 0x70, 0x2C, 0x0C, 0x78, 0xB7, 0x86, 0xA3, 0xF6, 0x23,
    0xB7, 0x38, 0xF5, 0xED, 0xF9, 0xAF, 0x83, 0x53, 0x0F, 0xB3, 0xFC, 0x54, 0xFA, 0xA2, 0x1E, 0xB9, 0xCF, 0x13, 0x31,
    0xFD, 0x0F, 0x0D, 0xA9, 0x54, 0xF6, 0x87, 0xCB, 0x9E, 0x18, 0x27, 0x96, 0x97, 0x90, 0x0E, 0x53, 0xFB, 0x31, 0x7C,
    0x9C, 0xBC, 0xE4, 0x8E, 0x23, 0xD0, 0x53, 0x71, 0xEC, 0xC1, 0x59, 0x51, 0xB8, 0xF3, 0x64, 0x9D, 0x7C, 0xA3, 0x3E,
    0xD6, 0x8D, 0xC9, 0x04, 0x7E, 0x82, 0xC9, 0xBA, 0xAD, 0x97, 0x99, 0xD0, 0xD4, 0x58, 0xCB, 0x84, 0x7C, 0xA9, 0xFF,
    0xBE, 0x3C, 0x8A, 0x77, 0x52, 0x33, 0x55, 0x7D, 0xDE, 0x13, 0xA8, 0xB1, 0x40, 0x87, 0xCC, 0x1B, 0xC8, 0xF1, 0x0F,
    0x6E, 0xCD, 0xD0, 0x83, 0xA9, 0x59, 0xCF, 0xF8, 0x4A, 0x9D, 0x1D, 0x50, 0x75, 0x5E, 0x3E, 0x19, 0x18, 0x18, 0xAF,
    0x23, 0xE2, 0x29, 0x35, 0x58, 0x76, 0x6D, 0x2C, 0x07, 0xE2, 0x57, 0x12, 0xB2, 0xCA, 0x0B, 0x53, 0x5E, 0xD8, 0xF6,
    0xC5, 0x6C, 0xE7, 0x3D, 0x24, 0xBD, 0xD0, 0x29, 0x17, 0x71, 0x86, 0x1A, 0x54, 0xB4, 0xC2, 0x85, 0xA9, 0xA3, 0xDB,
    0x7A, 0xCA, 0x6D, 0x22, 0x4A, 0xEA, 0xCD, 0x62, 0x1D, 0xB9, 0xF2, 0xA2, 0x2E, 0xD1, 0xE9, 0xE1, 0x1D, 0x75, 0xBE,
    0xD7, 0xDC, 0x0E, 0xCB, 0x0A, 0x8E, 0x68, 0xA2, 0xFF, 0x12, 0x63, 0x40, 0x8D, 0xC8, 0x08, 0xDF, 0xFD, 0x16, 0x4B,
    0x11, 0x67, 0x74, 0xCD, 0x0B, 0x9B, 0x8D, 0x05, 0x41, 0x1E, 0xD6, 0x26, 0x2E, 0x42, 0x9B, 0xA4, 0x95, 0x67, 0x6B,
    0x83, 0x98, 0xDB, 0x2F, 0x35, 0xD3, 0xC1, 0xB9, 0xCE, 0xD5, 0x26, 0x36, 0xF2, 0x76, 0x5E, 0x1A, 0x95, 0xCB, 0x7C,
    0xA4, 0xC3, 0xDD, 0xAB, 0xDD, 0xBF, 0xF3, 0x82, 0x53
]
# fmt: on

defender_def = """
/* ======== Generic Windows ======== */
/* https://learn.microsoft.com/en-us/windows/win32/api/winbase/ns-winbase-win32_stream_id */

enum STREAM_ID {
    DATA                 = 0x00000001,
    EA_DATA              = 0x00000002,
    SECURITY_DATA        = 0x00000003,
    ALTERNATE_DATA       = 0x00000004,
    LINK                 = 0x00000005,
    PROPERTY_DATA        = 0x00000006,
    OBJECT_ID            = 0x00000007,
    REPARSE_DATA         = 0x00000008,
    SPARSE_BLOCK         = 0x00000009,
    TXFS_DATA            = 0x0000000A,
    GHOSTED_FILE_EXTENTS = 0x0000000B,
};

flag STREAM_ATTRIBUTES {
    STREAM_NORMAL_ATTRIBUTE                 = 0x00000000,
    STREAM_MODIFIED_WHEN_READ               = 0x00000001,
    STREAM_CONTAINS_SECURITY                = 0x00000002,
    STREAM_CONTAINS_PROPERTIES              = 0x00000004,
    STREAM_SPARSE_ATTRIBUTE                 = 0x00000008,
    STREAM_CONTAINS_GHOSTED_FILE_EXTENTS    = 0x00000010,
};

typedef struct _WIN32_STREAM_ID {
    STREAM_ID           StreamId;
    STREAM_ATTRIBUTES   StreamAttributes;
    QWORD               Size;
    DWORD               StreamNameSize;
    WCHAR               StreamName[StreamNameSize / 2];
} WIN32_STREAM_ID;

/* ======== Defender Specific ======== */

enum FIELD_IDENTIFIER : WORD {
    CQuaResDataID_File      = 0x02,
    CQuaResDataID_Registry  = 0x03,
    Flags                   = 0x0A,
    PhysicalPath            = 0x0C,
    DetectionContext        = 0x0D,
    Unknown                 = 0x0E,
    CreationTime            = 0x0F,
    LastAccessTime          = 0x10,
    LastWriteTime           = 0x11
};

enum FIELD_TYPE : WORD {
    STRING          = 0x1,
    WSTRING         = 0x2,
    DWORD           = 0x3,
    RESOURCE_DATA   = 0x4,
    BYTES           = 0x5,
    QWORD           = 0x6,
};

struct QuarantineEntryFileHeader {
    CHAR        MagicHeader[4];
    CHAR        Unknown[4];
    CHAR        _Padding[32];
    DWORD       Section1Size;
    DWORD       Section2Size;
    DWORD       Section1CRC;
    DWORD       Section2CRC;
    CHAR        MagicFooter[4];
};

struct QuarantineEntrySection1 {
    CHAR    Id[16];
    CHAR    ScanId[16];
    QWORD   Timestamp;
    QWORD   ThreatId;
    DWORD   One;
    CHAR    DetectionName[];
};

struct QuarantineEntrySection2 {
    DWORD   EntryCount;
    DWORD   EntryOffsets[EntryCount];
};

struct QuarantineEntryResource {
    WCHAR   DetectionPath[];
    WORD    FieldCount;
    CHAR    DetectionType[];
};

struct QuarantineEntryResourceField {
    WORD        Size;
    WORD        Identifier:12;
    FIELD_TYPE  Type:4;
    CHAR        Data[Size];
};
"""

c_defender = cstruct().load(defender_def)

STREAM_ID = c_defender.STREAM_ID
STREAM_ATTRIBUTES = c_defender.STREAM_ATTRIBUTES
FIELD_IDENTIFIER = c_defender.FIELD_IDENTIFIER


def rc4_crypt(data: bytes) -> bytes:
    """RC4 encrypt / decrypt using the Defender Quarantine RC4 Key."""
    sbox = list(range(256))
    j = 0
    for i in range(256):
        j = (j + sbox[i] + DEFENDER_QUARANTINE_RC4_KEY[i]) % 256
        tmp = sbox[i]
        sbox[i] = sbox[j]
        sbox[j] = tmp

    out = bytearray(len(data))
    i = 0
    j = 0
    for k in range(len(data)):
        i = (i + 1) % 256
        j = (j + sbox[i]) % 256
        tmp = sbox[i]
        sbox[i] = sbox[j]
        sbox[j] = tmp
        val = sbox[(sbox[i] + sbox[j]) % 256]
        out[k] = val ^ data[k]

    return bytes(out)


def recover_quarantined_file_streams(fh: BinaryIO, filename: str) -> Iterator[tuple[str, bytes]]:
    """Recover the various data streams present in a quarantined file.

    Yields tuples of the output filename and the corresponding output data.
    """

    buf = BytesIO(rc4_crypt(fh.read()))

    while True:
        try:
            stream = c_defender.WIN32_STREAM_ID(buf)
        except EOFError:
            break
        data = buf.read(stream.Size)

        if stream.StreamId == STREAM_ID.DATA:
            yield (filename, data)
        elif stream.StreamId == STREAM_ID.EA_DATA:
            yield (f"{filename}.ea_data", data)
        elif stream.StreamId == STREAM_ID.SECURITY_DATA:
            yield (f"{filename}.security_descriptor", data)
        elif stream.StreamId == STREAM_ID.ALTERNATE_DATA:
            sanitized_stream_name = "".join(x for x in stream.StreamName if x.isalnum())
            yield (f"{filename}.{sanitized_stream_name}", data)
        elif stream.StreamId == STREAM_ID.LINK:
            yield (f"{filename}.link", data)
        elif stream.StreamId == STREAM_ID.PROPERTY_DATA:
            yield (f"{filename}.property_data", data)
        elif stream.StreamId == STREAM_ID.OBJECT_ID:
            yield (f"{filename}.object_id", data)
        elif stream.StreamId == STREAM_ID.REPARSE_DATA:
            yield (f"{filename}.reparse_data", data)
        elif stream.StreamId == STREAM_ID.SPARSE_BLOCK:
            yield (f"{filename}.sparse_block", data)
        elif stream.StreamId == STREAM_ID.TXFS_DATA:
            yield (f"{filename}.txfs_data", data)
        elif stream.StreamId == STREAM_ID.GHOSTED_FILE_EXTENTS:
            yield (f"{filename}.ghosted_file_extents", data)
        else:
            raise ValueError(f"Unexpected Stream ID {stream.StreamId}")


class QuarantineEntry:
    def __init__(self, fh: BinaryIO):
        # Decrypt & Parse the header so that we know the section sizes
        self.header = c_defender.QuarantineEntryFileHeader(rc4_crypt(fh.read(60)))

        # Decrypt & Parse Section 1. This will tell us some information about this quarantine entry.
        # These properties are shared for all quarantine entry resources associated with this quarantine entry.
        self.metadata = c_defender.QuarantineEntrySection1(rc4_crypt(fh.read(self.header.Section1Size)))

        self.timestamp = ts.wintimestamp(self.metadata.Timestamp)
        self.quarantine_id = self.metadata.Id
        self.scan_id = self.metadata.ScanId
        self.threat_id = self.metadata.ThreatId
        self.detection_name = self.metadata.DetectionName

        # The second section contains the number of quarantine entry resources contained in this quarantine entry,
        # as well as their offsets. After that, the individal quarantine entry resources start.
        resource_buf = BytesIO(rc4_crypt(fh.read(self.header.Section2Size)))
        resource_info = c_defender.QuarantineEntrySection2(resource_buf)

        # List holding all quarantine entry resources that belong to this quarantine entry.
        self.resources: list[QuarantineEntryResource] = []

        for offset in resource_info.EntryOffsets:
            resource_buf.seek(offset)
            self.resources.append(QuarantineEntryResource(resource_buf))


class QuarantineEntryResource:
    def __init__(self, fh: BinaryIO):
        self.metadata = c_defender.QuarantineEntryResource(fh)
        self.detection_path = self.metadata.DetectionPath
        self.field_count = self.metadata.FieldCount
        self.detection_type = self.metadata.DetectionType

        # It is possible that certain fields miss from a QuarantineEntryResource even though we expect them. Thus, we
        # initialize them in advance with a None value.
        self.resource_id = None
        self.creation_time = None
        self.last_access_time = None
        self.last_write_time = None

        self.unknown_fields = []

        # As the fields are aligned, we need to parse them individually
        offset = fh.tell()
        for _ in range(self.field_count):
            # Align
            offset = (offset + 3) & 0xFFFFFFFC
            fh.seek(offset)
            # Parse
            field = c_defender.QuarantineEntryResourceField(fh)
            self._add_field(field)

            # Move pointer
            offset += 4 + field.Size

    def _add_field(self, field: c_defender.QuarantineEntryResourceField) -> None:
        if field.Identifier == FIELD_IDENTIFIER.CQuaResDataID_File:
            self.resource_id = field.Data.hex().upper()
        elif field.Identifier == FIELD_IDENTIFIER.PhysicalPath:
            # Decoding as utf-16 leaves a trailing null-byte that we have to strip off.
            self.detection_path = field.Data.decode("utf-16").rstrip("\x00")
        elif field.Identifier == FIELD_IDENTIFIER.CreationTime:
            self.creation_time = ts.wintimestamp(int.from_bytes(field.Data, "little"))
        elif field.Identifier == FIELD_IDENTIFIER.LastAccessTime:
            self.last_access_time = ts.wintimestamp(int.from_bytes(field.Data, "little"))
        elif field.Identifier == FIELD_IDENTIFIER.LastWriteTime:
            self.last_write_time = ts.wintimestamp(int.from_bytes(field.Data, "little"))
        elif field.Identifier not in FIELD_IDENTIFIER:
            self.unknown_fields.append(field)



from __future__ import annotations

import io
from typing import TYPE_CHECKING, BinaryIO

from dissect.hypervisor import hdd

from dissect.target.container import Container

if TYPE_CHECKING:
    from pathlib import Path


class HddContainer(Container):
    __type__ = "hdd"

    def __init__(self, fh: Path, *args, **kwargs):
        if hasattr(fh, "read"):
            raise TypeError("HddContainer can only be opened by path")

        self.hdd = hdd.HDD(fh)
        self.stream = self.hdd.open()
        super().__init__(fh, self.stream.size, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return False

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        return path.suffix.lower() == ".hdd"

    def read(self, length: int) -> bytes:
        return self.stream.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.stream.seek(offset, whence)

    def tell(self) -> int:
        return self.stream.tell()

    def close(self) -> None:
        self.stream.close()



from __future__ import annotations

import ctypes
import platform
import re
import urllib.parse
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

from dissect.util.stream import BufferedStream

from dissect.target import filesystem, volume
from dissect.target.containers.raw import RawContainer
from dissect.target.exceptions import LoaderError
from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from collections.abc import Iterator
    from logging import Logger

    from dissect.target.target import Target

SOLARIS_DEV_DIR = Path("/dev/dsk")
SOLARIS_DRIVE_REGEX = re.compile(r".+d\d+$")

LINUX_DEV_DIR = Path("/dev")
LINUX_DRIVE_REGEX = re.compile(r"(([sh]|xv)d[a-z]$)|(fd\d+$)|(nvme\d+n\d+$)")
VOLATILE_LINUX_PATHS = [
    Path("/proc"),
    Path("/sys"),
]

ESXI_DEV_DIR = Path("/vmfs/devices/disks")

WINDOWS_ERROR_INSUFFICIENT_BUFFER = 0x7A
WINDOWS_DRIVE_FIXED = 3


class LocalLoader(Loader):
    """Load local filesystem."""

    def __init__(self, path: Path, **kwargs):
        kwargs["parsed_path"] = urllib.parse.urlparse(str(path))
        super().__init__(path, **kwargs, resolve=False)

    @staticmethod
    def detect(path: Path) -> bool:
        return urllib.parse.urlparse(str(path)).path == "local"

    def map(self, target: Target) -> None:
        # For the local loader we abuse the path/URI parsing a bit, so fix it up here
        target.parsed_path = urllib.parse.urlparse(str(target.path))
        target.path_query = self.parsed_query
        target.path = Path("local")

        os_name = _get_os_name()

        force_dirfs = "force-directory-fs" in self.parsed_query
        fallback_to_dirfs = "fallback-to-directory-fs" in self.parsed_query

        if os_name == "windows":
            map_windows_mounted_drives(target, force_dirfs=force_dirfs, fallback_to_dirfs=fallback_to_dirfs)
        else:
            if fallback_to_dirfs or force_dirfs:
                # Where Windows does some sophisticated fallback, for other
                # operating systems we don't know anything yet about the
                # relation between disks and mount points.
                # Temporary solution until we support parsing of /proc/mounts
                # and can implement OS-specific DirectoryFS fallback /
                # enforcement.
                target.filesystems.add(DirectoryFilesystem(Path("/")))
            elif os_name == "linux":
                map_linux_drives(target)
            elif os_name == "sunos":
                map_solaris_drives(target)
            elif os_name == "vmkernel":
                map_esxi_drives(target)
            elif os_name in ["darwin", "osx", "macos"]:
                # There is currently no way to access raw disk devices in OS-X,
                # so we always do a simple DirectoryFilesystem fallback.
                target.filesystems.add(DirectoryFilesystem(Path("/")))
            else:
                raise LoaderError(f"Unsupported OS for local target: {os_name}")


def map_linux_drives(target: Target) -> None:
    """Map Linux raw disks and /proc and /sys.

    Iterate through /dev and match raw device names (not partitions).

    /proc and /sys are mounted if they exists, allowing access to volatile files.
    """
    for drive in LINUX_DEV_DIR.iterdir():
        if LINUX_DRIVE_REGEX.match(drive.name):
            _add_disk_as_raw_container_to_target(drive, target)

    # Volatile filesystems are not present when running on a local target's raw
    # disks, so they are explicitly mounted here.
    #
    # Note that when running on a local target using a directory fs (through
    # force-directory-fs or fallback-to-directory-fs), these filesystems are
    # already present as they are usually mounted on the local system.
    for volatile_path in VOLATILE_LINUX_PATHS:
        if volatile_path.exists():
            volatile_fs = DirectoryFilesystem(volatile_path)
            target.filesystems.add(volatile_fs)
            target.fs.mount(str(volatile_path), volatile_fs)


def map_solaris_drives(target: Target) -> None:
    """Map Solaris raw disks.

    Iterate through /dev/dsk and match raw device names (not slices or partitions).
    """
    for drive in SOLARIS_DEV_DIR.iterdir():
        if not SOLARIS_DRIVE_REGEX.match(drive.name):
            continue
        _add_disk_as_raw_container_to_target(drive, target)


def map_esxi_drives(target: Target) -> None:
    """Map ESXi raw disks.

    Get all devices from /vmfs/devices/disks/* (not partitions).
    """
    for drive in ESXI_DEV_DIR.glob("vml.*"):
        if ":" in drive.name:
            continue
        _add_disk_as_raw_container_to_target(drive, target)


def map_windows_drives(target: Target) -> None:
    """Map Windows drives by iterating physical drives.

    For each physical drive, load the partition table and volumes.
    If a drive is encrypted using Bitlocker, use the OS transparent
    device to access it instead.

    Using this method, we get the drive serial and partition offset (MBR),
    or partition GUID (GPT), which we need for regular drive mounting.

    With this method we should open every partition of every disk,
    instead of only mounted drives.
    """
    devices = _windows_get_devices()

    physicaldrives = [d for d in devices if d.startswith("PhysicalDrive")]
    for drive in physicaldrives:
        drivenum = drive.replace("PhysicalDrive", "")
        drivepath = f"\\\\.\\{drive}"
        drivesize = None

        try:
            drivesize = _windows_get_disk_size(drivepath)
        except Exception as e:
            target.log.debug("Error getting size for drive %s", drive, exc_info=e)

        disk = RawContainer(
            BufferedStream(
                open(drivepath, "rb"),  # noqa: PTH123, SIM115
                size=drivesize,
            )
        )

        disk.vs = volume.open(disk)
        for i, vol in enumerate(disk.vs.volumes):
            # Open the decrypted volume if we encounter Bitlocker
            if b"-FVE-FS-" in vol.read(512):
                # Partition numbers are 1 indexed
                partname = f"Harddisk{drivenum}Partition{i + 1}"
                vol.fh = BufferedStream(
                    open(f"\\\\.\\{partname}", "rb"),  # noqa: PTH123, SIM115
                    size=vol.size,
                )
            vol.seek(0)
        disk.seek(0)

        # Alternative solution is to match partition by offset/guid
        # https://gist.github.com/santa4nt/11068180
        # https://docs.microsoft.com/en-us/windows/desktop/api/fltuser/nf-fltuser-filtergetdosname

        target.disks.add(disk)


def _add_disk_as_raw_container_to_target(drive: Path, target: Target) -> None:
    try:
        fh = BufferedStream(drive.open("rb"))
        target.disks.add(RawContainer(fh))
    except Exception as e:
        target.log.warning("Unable to open drive: %s, skipped", drive)
        target.log.debug("", exc_info=e)


def _read_drive_letters() -> list[bytes]:
    # Get all logical drive letters
    drives_buf = ctypes.c_buffer(256)
    ctypes.windll.kernel32.GetLogicalDriveStringsA(256, drives_buf)

    return drives_buf.raw.rstrip(b"\x00").split(b"\x00")


def _get_windows_drive_volumes(log: Logger) -> Iterator[tuple[volume.Volume, bool, RawContainer | None, int | None]]:
    # Get the sysvol drive letter
    windir_buf = ctypes.c_buffer(256)
    ctypes.windll.kernel32.GetWindowsDirectoryA(windir_buf, 256)
    windrive = windir_buf.raw[:2].decode().lower()

    disk_map = {}

    for drive_letter in _read_drive_letters():
        drive_letter = drive_letter.rstrip(b"\\")
        # Check for fixed drives
        drive_type = ctypes.windll.kernel32.GetDriveTypeA(drive_letter)
        if drive_type != WINDOWS_DRIVE_FIXED:
            continue

        drive_letter = drive_letter.decode().lower()
        drive_path = f"\\\\.\\{drive_letter}"

        # Attempt to read from drive
        # If it fails, skip it
        try:
            log.debug("Trying to read from drive %s", drive_letter)
            drive_volume = volume.Volume(BufferedStream(open(drive_path, "rb")), None, 0, None, None, None)  # noqa: PTH123, SIM115
            drive_volume.seek(0)
            drive_volume.read(1024)
            drive_volume.seek(0)
        except Exception as e:
            log.debug("Error reading drive %s", drive_letter, exc_info=e)
            continue

        drive_volume.drive_letter = drive_letter

        try:
            extent_info = _windows_get_volume_disk_extents(drive_path)
        except Exception as e:
            log.debug("Failed to get volume disk extents: %s", drive_path, exc_info=e)
            extent_info = None

        if extent_info and extent_info.NumberOfDiskExtents == 1:
            extent = extent_info.Extents[0]
            disk_num = extent.DiskNumber

            # Set offset and size now that we know them
            drive_volume.offset = extent.StartingOffset
            drive_volume.size = extent.ExtentLength
            drive_volume.fh.size = extent.ExtentLength

            if disk_num not in disk_map:
                disk_path = f"\\\\.\\PhysicalDrive{disk_num}"

                # Check if drive can be accessed (skip emulated drives like RAM disks)
                if not _is_physical_drive(disk_path):
                    log.debug(
                        "Skipped drive %d from %s, not a physical drive (could be emulation or ram disk)",
                        disk_num,
                        drive_path,
                    )
                    continue

                try:
                    disk_size = _windows_get_disk_size(disk_path)
                except Exception as e:
                    log.debug("Error getting size for disk %s", disk_path, exc_info=e)
                    disk_size = None
                try:
                    disk = RawContainer(
                        BufferedStream(
                            open(disk_path, "rb"),  # noqa: PTH123, SIM115
                            size=disk_size,
                        )
                    )
                    disk_map[disk_num] = disk
                except Exception as e:
                    log.debug("Unable to open disk %d at %s, skipped", disk_num, disk_path, exc_info=e)
                    continue
                try:
                    disk.vs = volume.open(disk)
                except Exception as e:
                    log.debug("Failed to open volume system on disk %s", disk_path, exc_info=e)
                disk.seek(0)
            else:
                disk = disk_map[disk_num]

            if disk.vs:
                # Find the matching volume on the disk
                for v in disk.vs.volumes:
                    if v.offset == drive_volume.offset:
                        v.fh = drive_volume.fh
                        drive_volume = v
                        break
            else:
                # Opening a volume system failed, map the disk to the volume anyway
                drive_volume.disk = disk
        else:
            disk = None
            disk_num = None

        yield (drive_volume, drive_letter == windrive, disk, disk_num)


def map_windows_mounted_drives(target: Target, force_dirfs: bool = False, fallback_to_dirfs: bool = False) -> None:
    """Map Windows drives by their drive letter.

    For each drive (mounted) partition, determine if it's a fixed drive
    and if it's readable. If it is, add it as a volume to the target.

    Since we don't know the drive serial and other information, we
    already mount filesystems to drive letters (which we do know).

    Downside to this method is that we only open mounted volumes.
    Upside is that we can also open BDE/LDM/Storage space volumes.

    Some inspiration drawn from http://velisthoughts.blogspot.com/2012/02/enumerating-and-using-partitions-and.html
    """

    disks = {}
    for drive_volume, is_windrive, disk, disk_num in _get_windows_drive_volumes(target.log):
        if disk:
            disks[disk_num] = disk

        if not disk or not disk.vs:
            target.volumes.add(drive_volume)

        if is_windrive:
            continue

        if force_dirfs:
            drive_volume.fs = DirectoryFilesystem(Path(f"{drive_volume.drive_letter}:\\"))
        else:
            try:
                drive_volume.fs = filesystem.open(drive_volume)
            except Exception as e:  # a wide exception to catch all issues with the filesystem
                target.log.debug(
                    "Failed to open filesystem of drive %s (volume: %s)",
                    drive_volume.drive_letter,
                    drive_volume,
                    exc_info=e,
                )
                if fallback_to_dirfs:
                    drive_volume.fs = DirectoryFilesystem(Path(f"{drive_volume.drive_letter}:\\"))

        if drive_volume.fs and drive_volume.drive_letter:
            target.fs.mount(drive_volume.drive_letter, drive_volume.fs)

    # Add all the disks we found.
    # DiskCollection.apply() will add the volumes found within them.
    for disk_num in sorted(disks.keys()):
        target.disks.add(disks[disk_num])


@cache
def _windows_get_devices() -> list[str]:
    """Internal function to query all devices.

    https://www.virag.si/2010/02/enumerate-physical-drives-in-windows/
    """
    bufsize = 256
    devices_buf = ctypes.c_buffer(bufsize)
    return_size = 0

    while return_size == 0:
        return_size = ctypes.windll.kernel32.QueryDosDeviceA(None, devices_buf, bufsize)

        if return_size == 0:
            errcode = ctypes.windll.kernel32.GetLastError()
            if errcode == WINDOWS_ERROR_INSUFFICIENT_BUFFER:
                bufsize *= 2
                devices_buf = ctypes.c_buffer(bufsize)
                continue
            raise ValueError

    return [d.decode() for d in devices_buf.raw.rstrip(b"\x00").split(b"\x00")]


def _is_physical_drive(path: str) -> bool:
    path = path.replace("\\\\.\\", "")
    return path in _windows_get_devices()


def _windows_get_disk_size(path: str) -> int:
    geometry_ex = _windows_get_disk_geometry_ex(path)
    return geometry_ex.DiskSize


def _windows_get_disk_geometry_ex(path: str) -> ctypes.Structure:
    from ctypes import wintypes

    IOCTL_DISK_GET_DRIVE_GEOMETRY_EX = 0x700A0

    class DISK_GEOMETRY(ctypes.Structure):
        _fields_ = (
            ("Cylinders", wintypes.LARGE_INTEGER),
            ("MediaType", wintypes.BYTE),
            ("TracksPerCylinder", wintypes.DWORD),
            ("SectorsPerTrack", wintypes.DWORD),
            ("BytesPerSector", wintypes.DWORD),
        )

    class DISK_GEOMETRY_EX(ctypes.Structure):
        _fields_ = (
            ("Geometry", DISK_GEOMETRY),
            ("DiskSize", wintypes.LARGE_INTEGER),
            ("Data", wintypes.BYTE),
        )

    handle = _windows_createfile(path)
    try:
        status, res = _windows_ioctl(handle, IOCTL_DISK_GET_DRIVE_GEOMETRY_EX, DISK_GEOMETRY_EX)
    finally:
        _windows_closehandle(handle)

    if status == 0:
        err = ctypes.windll.kernel32.GetLastError()
        raise OSError(f"unable to execute IOCTL_DISK_GET_DRIVE_GEOMETRY_EX, error: 0x{err:08x}")

    return res


def _windows_get_volume_disk_extents(path: str) -> ctypes.Structure:
    from ctypes import wintypes

    ERROR_MORE_DATA = 234
    IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS = 0x560000

    class DISK_EXTENT(ctypes.Structure):
        _fields_ = (
            ("DiskNumber", wintypes.DWORD),
            ("StartingOffset", wintypes.LARGE_INTEGER),
            ("ExtentLength", wintypes.LARGE_INTEGER),
        )

    class VOLUME_DISK_EXTENTS(ctypes.Structure):
        _fields_ = (
            ("NumberOfDiskExtents", wintypes.DWORD),
            ("Extents", DISK_EXTENT * 1),
        )

    handle = _windows_createfile(path)
    try:
        status, res = _windows_ioctl(handle, IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS, VOLUME_DISK_EXTENTS)
    finally:
        _windows_closehandle(handle)

    if status == 0:
        err = ctypes.windll.kernel32.GetLastError()
        if err != ERROR_MORE_DATA:
            raise OSError(f"unable to execute IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS, error: 0x{err:08x}")

    return res


def _windows_createfile(path: str) -> int:
    OPEN_EXISTING = 3
    FILE_ATTRIBUTE_NORMAL = 0x00000080

    handle = ctypes.windll.kernel32.CreateFileW(
        path,
        0,
        0,
        0,
        OPEN_EXISTING,
        FILE_ATTRIBUTE_NORMAL,
        0,
    )

    if handle == -1:
        err = ctypes.windll.kernel32.GetLastError()
        raise OSError(f"unable to open handle to {path}, error: 0x{err:08x}")

    return handle


def _windows_closehandle(handle: int) -> None:
    ctypes.windll.kernel32.CloseHandle(handle)


T = TypeVar("T", bound=ctypes.Structure)


def _windows_ioctl(handle: int, ioctl: int, out_struct: type[T]) -> tuple[int, T]:
    # http://www.ioctls.net/
    from ctypes import wintypes

    out_inst = out_struct()
    bytes_returned = wintypes.DWORD(0)
    status = ctypes.windll.kernel32.DeviceIoControl(
        handle,
        ioctl,
        None,
        0,
        ctypes.pointer(out_inst),
        ctypes.sizeof(out_struct),
        ctypes.byref(bytes_returned),
        None,
    )

    return status, out_inst


def _get_os_name() -> str:
    return platform.system().lower()



from __future__ import annotations

import logging
import re
from datetime import datetime
from typing import TYPE_CHECKING

from dissect.target.helpers.fsutil import open_decompress

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

log = logging.getLogger(__name__)

RE_TS = re.compile(r"^[A-Za-z]{3}\s*\d{1,2}\s\d{1,2}:\d{2}:\d{2}")
RE_TS_ISO = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}\+\d{2}:\d{2}")
RE_LINE = re.compile(
    r"""
    \d{2}:\d{2}\s                           # First match on the similar ending of the different timestamps
    ((?:\S+)\s)?                            # The hostname (optional), but do not capture it
    (?P<service>\S+?)(\[(?P<pid>\d+?)\])?:  # The service / daemon with optionally the PID between brackets
    (\s*(?P<message>.+?)\s*)?$              # The log message stripped from spaces left and right
    """,
    re.VERBOSE,
)


def iso_readlines(file: Path, max_lines: int | None = None) -> Iterator[tuple[datetime, str]]:
    """Iterator reading the provided log file in ISO format. Mimics ``year_rollover_helper`` behaviour."""
    with open_decompress(file, "rt") as fh:
        for i, line in enumerate(fh):
            if max_lines is not None and i >= max_lines:
                log.debug("Stopping iso_readlines enumeration in %s: max_lines=%s was reached", file, max_lines)
                break

            if not (match := RE_TS_ISO.match(line)):
                if not max_lines:
                    log.warning("No timestamp found in one of the lines in %s!", file)
                log.debug("Skipping line: %s", line)
                continue

            try:
                ts = datetime.strptime(match[0], "%Y-%m-%dT%H:%M:%S.%f%z")
            except ValueError as e:
                log.warning("Unable to parse ISO timestamp in line: %s", line)
                log.debug("", exc_info=e)
                continue

            yield ts, line


def is_iso_fmt(file: Path) -> bool:
    """Determine if the provided log file uses ISO 8601 timestamp format logging or not."""
    # We do not want to iterate of the entire file so we limit iso_readlines to the first few lines.
    # We can not use islice here since that would only work if the file is ISO formatted and thus yields results.
    return any(iso_readlines(file, max_lines=3))



from __future__ import annotations

import logging
import sys
from typing import Any

import structlog


def custom_obj_renderer(
    logger: structlog.types.WrappedLogger,
    name: str,
    event_dict: structlog.types.EventDict,
) -> dict[Any, str]:
    """Simple ``str()`` serialization for the event dictionary values for purely aesthetic reasons."""
    return {key: str(value) for key, value in event_dict.items()}


def render_stacktrace_only_in_debug_or_less(
    logger: structlog.types.WrappedLogger,
    name: str,
    event_dict: structlog.types.EventDict,
) -> dict[Any, str]:
    """Render a stack trace of an exception only if ``logger`` is configured with ``DEBUG`` or lower level,
    otherwise render ``str()`` representation of an exception.
    """
    # If configured logging level is less permissive than DEBUG,
    # do not render full stack trace
    # https://docs.python.org/3/library/logging.html#logging-levels
    if event_dict.get("exc_info") and logger.getEffectiveLevel() > logging.DEBUG:
        event_dict.pop("exc_info")
        _, exc, _ = sys.exc_info()
        event_dict["exc"] = str(exc)
    return event_dict


def configure_logging(verbose_value: int, be_quiet: bool, as_plain_text: bool = True) -> None:
    """Configure logging level for ``dissect`` root logger.

    By default, if ``verbose_value`` is not set (equals ``0``) and ``be_quiet`` is ``False``,
    set logging level for ``dissect`` root logger to ``WARNING``.

    If ``be_quiet`` is set to ``True``, logging level is set to the least noisy ``CRITICAL`` level.
    """

    renderer = (
        structlog.dev.ConsoleRenderer(colors=True, pad_event=10)
        if as_plain_text
        else structlog.processors.JSONRenderer(sort_keys=True)
    )

    attr_processors = [
        # Add the name of the logger to event dict.
        structlog.stdlib.add_logger_name,
        # Add log level to event dict.
        structlog.stdlib.add_log_level,
        # Add a timestamp in ISO 8601 format.
        structlog.processors.TimeStamper(fmt="iso"),
    ]

    structlog.configure(
        processors=(
            [
                # If log level is too low, abort pipeline and throw away log entry.
                structlog.stdlib.filter_by_level,
                *attr_processors,
                # Perform %-style formatting.
                structlog.stdlib.PositionalArgumentsFormatter(),
                # If the "stack_info" key in the event dict is true, remove it and
                # render the current stack trace in the "stack" key.
                structlog.processors.StackInfoRenderer(),
                custom_obj_renderer,
                render_stacktrace_only_in_debug_or_less,
                # Wrapping is needed in order to use formatter down the line
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
            ]
        ),
        # `wrapper_class` is the bound logger that you get back from
        # get_logger(). This one imitates the API of `logging.Logger`.
        wrapper_class=structlog.stdlib.BoundLogger,
        # `logger_factory` is used to create wrapped loggers that are used for
        # OUTPUT. This one returns a `logging.Logger`.
        logger_factory=structlog.stdlib.LoggerFactory(),
        # Effectively freeze configuration after creating the first bound
        # logger.
        cache_logger_on_first_use=True,
    )

    # Warnings issued by the ``warnings`` module will be
    # redirected to the ``py.warnings`` logger
    logging.captureWarnings(True)

    dissect_logger = logging.getLogger("dissect")

    if be_quiet:
        dissect_logger.setLevel(level=logging.CRITICAL)
    elif verbose_value == 0:
        dissect_logger.setLevel(level=logging.WARNING)
    elif verbose_value == 1:
        dissect_logger.setLevel(level=logging.INFO)
    elif verbose_value > 1:
        dissect_logger.setLevel(level=logging.DEBUG)
    else:
        pass

    formatter = structlog.stdlib.ProcessorFormatter(processor=renderer, foreign_pre_chain=attr_processors)

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    # Check if we need to cleanup old logging handlers
    root_logger = logging.getLogger()
    for handler in root_logger.handlers[:]:
        if isinstance(handler, logging.StreamHandler) and isinstance(
            handler.formatter, structlog.stdlib.ProcessorFormatter
        ):
            # We already configured logging once, assume we want to reconfigure
            root_logger.removeHandler(handler)
            handler.close()

    # Add the new handler to the root logger
    root_logger.addHandler(handler)



from __future__ import annotations

from operator import attrgetter
from typing import TYPE_CHECKING, Any, Callable, Union, get_args

from flow.record.fieldtypes.net import IPAddress, IPNetwork

from dissect.target.helpers.record import (
    MacOSInterfaceRecord,
    UnixInterfaceRecord,
    WindowsInterfaceRecord,
)
from dissect.target.plugin import Plugin, export, internal

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

InterfaceRecord = Union[UnixInterfaceRecord, WindowsInterfaceRecord, MacOSInterfaceRecord]


class NetworkPlugin(Plugin):
    """Generic implementation for network interfaces plugin."""

    __namespace__ = "network"

    def __init__(self, target: Target):
        super().__init__(target)
        self._interface_list: list[InterfaceRecord] | None = None

    def check_compatible(self) -> None:
        pass

    def _interfaces(self) -> Iterator[InterfaceRecord]:
        yield from ()

    def _get_record_type(self, field_name: str, func: Callable[[Any], Any] | None = None) -> Iterator[Any]:
        for record in self.interfaces():
            if (output := getattr(record, field_name, None)) is None:
                continue

            if not isinstance(output, list):
                output = [output]

            yield from (map(func, output) if func else output)

    @export(record=get_args(InterfaceRecord))
    def interfaces(self) -> Iterator[InterfaceRecord]:
        """Yield interfaces."""
        # Only search for the interfaces once
        if self._interface_list is None:
            self._interface_list = list(self._interfaces())

        yield from self._interface_list

    @export
    def ips(self) -> list[IPAddress]:
        """Return IP addresses as list of :class:`IPAddress`."""
        return list(set(self._get_record_type("cidr", attrgetter("ip"))))

    @export
    def gateways(self) -> list[IPAddress]:
        """Return gateways as list of :class:`IPAddress`."""
        return list(set(self._get_record_type("gateway")))

    @export
    def macs(self) -> list[str]:
        """Return MAC addresses as list of :class:`str`."""
        return list(set(self._get_record_type("mac")))

    @export
    def dns(self) -> list[str | IPAddress]:
        """Return DNS addresses as list of :class:`str`."""
        return list(set(self._get_record_type("dns")))

    @internal
    def with_ip(self, ip_addr: str) -> Iterator[InterfaceRecord]:
        """Yield all interfaces with the given IP address."""
        for interface in self.interfaces():
            if any(iface.ip == ip_addr for iface in interface.cidr):
                yield interface

    @internal
    def with_mac(self, mac: str) -> Iterator[InterfaceRecord]:
        """Yield all interfaces with the given full or partial MAC address."""
        for interface in self.interfaces():
            if mac in interface.mac:
                yield interface

    @internal
    def in_cidr(self, cidr: str) -> Iterator[InterfaceRecord]:
        """Yield all interfaces with IP addresses in the given CIDR range."""
        cidr = IPNetwork(cidr)
        for interface in self.interfaces():
            if any(iface.ip in cidr for iface in interface.cidr):
                yield interface



from __future__ import annotations

import logging
import lzma
from typing import TYPE_CHECKING, Any, BinaryIO, Callable

import zstandard
from dissect.cstruct import cstruct
from dissect.util import ts
from dissect.util.compression import lz4

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.helpers.utils import IntEnumMissing
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

log = logging.getLogger(__name__)

# The events have undocumented fields that are not part of the record
JournalRecord = TargetRecordDescriptor(
    "linux/log/journal",
    [
        ("datetime", "ts"),
        # User Journal fields
        ("string", "message"),
        ("string", "message_id"),
        ("string", "priority"),
        ("path", "code_file"),
        ("varint", "code_line"),
        ("string", "code_func"),
        ("varint", "errno"),
        ("string", "invocation_id"),
        ("string", "user_invocation_id"),
        ("string", "syslog_facility"),
        ("string", "syslog_identifier"),
        ("varint", "syslog_pid"),
        ("string", "syslog_raw"),
        ("string", "documentation"),
        ("varint", "tid"),
        ("string", "unit"),
        ("string", "user_unit"),
        # Trusted Journal fields
        ("varint", "pid"),
        ("varint", "uid"),
        ("varint", "gid"),
        ("string", "comm"),
        ("path", "exe"),
        ("string", "cmdline"),
        ("string", "cap_effective"),
        ("varint", "audit_session"),
        ("varint", "audit_loginuid"),
        ("path", "systemd_cgroup"),
        ("string", "systemd_slice"),
        ("string", "systemd_unit"),
        ("string", "systemd_user_unit"),
        ("string", "systemd_user_slice"),
        ("string", "systemd_session"),
        ("string", "systemd_owner_uid"),
        ("string", "selinux_context"),
        ("string", "kernel_boot_id"),
        ("string", "machine_id"),
        ("string", "systemd_invocation_id"),
        ("string", "transport"),
        ("string", "stream_id"),
        ("string", "line_break"),
        ("string", "namespace"),
        ("string", "runtime_scope"),
        # Kernel Journal fields
        ("string", "kernel_device"),
        ("string", "kernel_subsystem"),
        ("string", "udev_sysname"),
        ("path", "udev_devnode"),
        ("path", "udev_devlink"),
        # Other fields
        ("string", "journal_hostname"),
        ("path", "source"),
    ],
)

journal_def = """
#define HEADER_SIGNATURE b"LPKSHHRH"

typedef uint8 uint8_t;
typedef uint32 le32_t;
typedef uint64 le64_t;

enum State : uint8_t {
    OFFLINE   =  0,
    ONLINE    =  1,
    ARCHIVED  =  2,
    UNKNOWN
};

union sd_id128_t {
    uint8_t   bytes[16];
    uint64_t  qwords[2];
};

flag IncompatibleFlag : le32_t {
    HEADER_INCOMPATIBLE_COMPRESSED_XZ   = 1,
    HEADER_INCOMPATIBLE_COMPRESSED_LZ4  = 2,
    HEADER_INCOMPATIBLE_KEYED_HASH      = 4,
    HEADER_INCOMPATIBLE_COMPRESSED_ZSTD = 8,
    HEADER_INCOMPATIBLE_COMPACT         = 16,             // indicates that the Journal file uses the new binary format
};

struct Header {
    char              signature[8];
    le32_t            compatible_flags;
    IncompatibleFlag  incompatible_flags;
    State             state;
    uint8_t           reserved[7];
    sd_id128_t        file_id;
    sd_id128_t        machine_id;
    sd_id128_t        tail_entry_boot_id;
    sd_id128_t        seqnum_id;
    le64_t            header_size;
    le64_t            arena_size;
    le64_t            data_hash_table_offset;
    le64_t            data_hash_table_size;
    le64_t            field_hash_table_offset;
    le64_t            field_hash_table_size;
    le64_t            tail_object_offset;
    le64_t            n_objects;
    le64_t            n_entries;
    le64_t            tail_entry_seqnum;
    le64_t            head_entry_seqnum;
    le64_t            entry_array_offset;
    le64_t            head_entry_realtime;
    le64_t            tail_entry_realtime;
    le64_t            tail_entry_monotonic;
    le64_t            n_data;
    le64_t            n_fields;
    le64_t            n_tags;
    le64_t            n_entry_arrays;
    le64_t            data_hash_chain_depth;
    le64_t            field_hash_chain_depth;
    le32_t            tail_entry_array_offset;
    le32_t            tail_entry_array_n_entries;
    le64_t            tail_entry_offset;
};

enum ObjectType : uint8 {
    OBJECT_UNUSED,
    OBJECT_DATA,
    OBJECT_FIELD,
    OBJECT_ENTRY,
    OBJECT_DATA_HASH_TABLE,
    OBJECT_FIELD_HASH_TABLE,
    OBJECT_ENTRY_ARRAY,
    OBJECT_TAG,
    _OBJECT_TYPE_MAX
};

flag ObjectFlag : uint8 {
    OBJECT_UNCOMPRESSED      =  0,
    OBJECT_COMPRESSED_XZ     =  1,
    OBJECT_COMPRESSED_LZ4    =  2,
    OBJECT_COMPRESSED_ZSTD   =  4,
    _OBJECT_COMPRESSED_MASK  =  7
};

struct ObjectHeader {
    ObjectType  type;                                     // The type field is one of the object types listed above
    uint8_t     flags;                                    // If DATA object the value is ObjectFlag
    uint8_t     reserved[6];
    le64_t      size;                                     // The size field encodes the size of the object including all its headers and payload
};


// The first four members are copied from ObjectHeader, so that the size can be used as the length of payload
struct DataObject {
    // ObjectType  type;
    ObjectFlag  flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      hash;
    le64_t      next_hash_offset;
    le64_t      next_field_offset;
    le64_t      entry_offset;
    le64_t      entry_array_offset;
    le64_t      n_entries;
    char        payload[size - 64];                       // Data objects carry actual field data in the payload[] array.
};

// If the HEADER_INCOMPATIBLE_COMPACT flag is set, two extra fields are stored to allow immediate access
// to the tail entry array in the DATA object's entry array chain.
struct DataObject_Compact {
    // ObjectType  type;
    ObjectFlag  flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      hash;
    le64_t      next_hash_offset;
    le64_t      next_field_offset;
    le64_t      entry_offset;
    le64_t      entry_array_offset;
    le64_t      n_entries;
    le32_t      tail_entry_array_offset;
    le32_t      tail_entry_array_n_entries;
    char        payload[size - 72];                       // Data objects carry actual field data in the payload[] array.
};

struct EntryItem {
    le64_t object_offset;
    le64_t hash;
};

struct EntryItem_Compact {
    le32_t object_offset;
}

// The first four members are copied from ObjectHeader, so that the size can be used as the length of items
struct EntryObject {
    ObjectType  type;
    uint8_t     flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      seqnum;
    le64_t      realtime;
    le64_t      monotonic;
    sd_id128_t  boot_id;
    le64_t      xor_hash;
    EntryItem   items[(size - 64) / 16];                  // The size minus the previous members divided by the size of the items
};

// If the HEADER_INCOMPATIBLE_COMPACT flag is set, DATA object offsets are stored as 32-bit integers instead of 64bit
// and the unused hash field per data object is not stored anymore.
struct EntryObject_Compact {
    ObjectType  type;
    uint8_t     flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      seqnum;
    le64_t      realtime;
    le64_t      monotonic;
    sd_id128_t  boot_id;
    le64_t      xor_hash;
    EntryItem_Compact   items[(size - 64) / 4];
};

// The first four members are copied from from ObjectHeader, so that the size can be used as the length of entry_object_offsets
struct EntryArrayObject {
    // ObjectType  type;
    uint8_t     flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      next_entry_array_offset;
    le64_t      entry_object_offsets[(size - 24) / 8];    // The size minus the previous members divided by the size of the offset
};

struct EntryArrayObject_Compact {
    // ObjectType  type;
    uint8_t     flags;
    uint8_t     reserved[6];
    le64_t      size;
    le64_t      next_entry_array_offset;
    le32_t      entry_object_offsets[(size - 24) / 4];
};
"""  # noqa: E501

c_journal = cstruct().load(journal_def)


class JournalMessagePriority(IntEnumMissing):
    """Journal message priority enum.

    References:
        - https://wiki.archlinux.org/title/Systemd/Journal
    """

    EMERG = 0
    ALERT = 1
    CRIT = 2
    ERR = 3
    WARNING = 4
    NOTICE = 5
    INFO = 6
    DEBUG = 7


def get_optional(value: str, to_type: Callable) -> Any | None:
    """Return the value if True, otherwise return None."""

    if not value:
        return None

    try:
        return to_type(value)

    except ValueError as e:
        log.error("Unable to cast '%s' to %s", value, to_type)  # noqa: TRY400
        log.debug("", exc_info=e)
        return None


# Sometimes stringy None is inserted by external tools like Ansible
def int_or_none(value: str) -> int | None:
    return int(value) if value and value != "None" else None


class JournalFile:
    """Parse Systemd Journal file format.

    References:
        - https://github.com/systemd/systemd/blob/206f0f397edf1144c63a158fb30f496c3e89f256/docs/JOURNAL_FILE_FORMAT.md
        - https://github.com/libyal/dtformats/blob/c4fc2b8102702c64b58f145971821986bf74e6c0/documentation/Systemd%20journal%20file%20format.asciidoc
    """

    def __init__(self, fh: BinaryIO, target: Target):
        self.fh = fh
        self.target = target

        try:
            self.header = c_journal.Header(self.fh)
        except EOFError as e:
            raise ValueError(f"Invalid systemd Journal file: {e}")

        if self.header.signature != c_journal.HEADER_SIGNATURE:
            raise ValueError(f"Journal file has invalid magic header: {self.header.signature!r}'")

    def decode_value(self, value: bytes) -> tuple[str, str]:
        """Decode the given bytes to a key value pair."""
        value = value.decode(errors="surrogateescape").strip().lstrip("_")
        key, value = value.split("=", 1)
        key = key.lower()
        return key, value

    def __iter__(self) -> Iterator[dict[str, int | str]]:
        "Iterate over the entry objects to read payloads."

        offset = self.header.entry_array_offset
        while offset != 0:
            self.fh.seek(offset)
            object_type = self.fh.read(1)[0]

            if object_type == c_journal.ObjectType.OBJECT_UNUSED:
                self.target.log.warning(
                    "ObjectType OBJECT_UNUSED encountered for next OBJECT_ENTRY_ARRAY offset at 0x%X. "
                    "This indicates allocated space in the journal file which is not used yet.",
                    offset,
                )
                break

            elif object_type != c_journal.ObjectType.OBJECT_ENTRY_ARRAY:
                raise ValueError(f"Expected OBJECT_ENTRY_ARRAY or OBJECT_UNUSED at offset {offset}")

            if self.header.incompatible_flags & c_journal.IncompatibleFlag.HEADER_INCOMPATIBLE_COMPACT:
                entry_array_object = c_journal.EntryArrayObject_Compact(self.fh)
            else:
                entry_array_object = c_journal.EntryArrayObject(self.fh)

            for entry_object_offset in entry_array_object.entry_object_offsets:
                if entry_object_offset:
                    yield from self._parse_entry_object(offset=entry_object_offset)

            offset = entry_array_object.next_entry_array_offset

    def _parse_entry_object(self, offset: int) -> Iterator[dict]:
        self.fh.seek(offset)

        try:
            if self.header.incompatible_flags & c_journal.IncompatibleFlag.HEADER_INCOMPATIBLE_COMPACT:
                entry = c_journal.EntryObject_Compact(self.fh)
            else:
                entry = c_journal.EntryObject(self.fh)

        except EOFError as e:
            self.target.log.warning("Unable to read Journal EntryObject at offset %s in: %s", offset, self.fh)
            self.target.log.debug("", exc_info=e)
            return

        event = {"ts": ts.from_unix_us(entry.realtime)}
        for item in entry.items:
            try:
                self.fh.seek(item.object_offset)

                if self.fh.read(1)[0] != c_journal.ObjectType.OBJECT_DATA:
                    continue

                if self.header.incompatible_flags & c_journal.IncompatibleFlag.HEADER_INCOMPATIBLE_COMPACT:
                    data_object = c_journal.DataObject_Compact(self.fh)
                else:
                    data_object = c_journal.DataObject(self.fh)

                if not data_object.payload:
                    continue

                data = data_object.payload

                if data_object.flags & c_journal.ObjectFlag.OBJECT_COMPRESSED_XZ:
                    data = lzma.decompress(data)

                elif data_object.flags & c_journal.ObjectFlag.OBJECT_COMPRESSED_LZ4:
                    data = lz4.decompress(data[8:])

                elif data_object.flags & c_journal.ObjectFlag.OBJECT_COMPRESSED_ZSTD:
                    data = zstandard.decompress(data)

                key, value = self.decode_value(data)
                event[key] = value

            except Exception as e:
                self.target.log.warning(
                    "Journal DataObject could not be parsed at offset %s in %s",
                    item.object_offset,
                    getattr(self.fh, "name", None),
                )
                self.target.log.debug("", exc_info=e)
                continue

        yield event


class JournalPlugin(Plugin):
    """Systemd Journal plugin."""

    JOURNAL_PATHS = ("/var/log/journal",)  # TODO: /run/systemd/journal
    JOURNAL_GLOB = "*/*.journal*"  # The extensions .journal and .journal~

    def __init__(self, target: Target):
        super().__init__(target)
        self.journal_files = []

        for journal_path in self.JOURNAL_PATHS:
            self.journal_files.extend(self.target.fs.path(journal_path).glob(self.JOURNAL_GLOB))

    def check_compatible(self) -> None:
        if not self.journal_files:
            raise UnsupportedPluginError("No journald files found")

    @export(record=JournalRecord)
    def journal(self) -> Iterator[JournalRecord]:
        """Return the contents of Systemd Journal log files.

        References:
            - https://wiki.archlinux.org/title/Systemd/Journal
            - https://github.com/systemd/systemd/blob/9203abf79f1d05fdef9b039e7addf9fc5a27752d/man/systemd.journal-fields.xml
        """
        path_function = self.target.fs.path

        for journal_file in self.journal_files:
            if not journal_file.is_file():
                self.target.log.warning("Unable to parse journal file as it is not a file: %s", journal_file)
                continue

            try:
                fh = journal_file.open()
                journal = JournalFile(fh, self.target)

            except Exception as e:
                self.target.log.warning("Unable to parse journal file structure: %s: %s", journal_file, str(e))
                self.target.log.debug("", exc_info=e)
                continue

            for entry in journal:
                priority = int_or_none(entry.get("priority"))
                yield JournalRecord(
                    ts=entry.get("ts"),
                    message=entry.get("message"),
                    message_id=entry.get("message_id"),
                    priority=JournalMessagePriority(priority).name.lower() if priority else None,
                    code_file=get_optional(entry.get("code_file"), path_function),
                    code_line=int_or_none(entry.get("code_line")),
                    code_func=entry.get("code_func"),
                    errno=int_or_none(entry.get("errno")),
                    invocation_id=entry.get("invocation_id"),
                    user_invocation_id=entry.get("user_invocation_id"),
                    syslog_facility=entry.get("syslog_facility"),
                    syslog_identifier=entry.get("syslog_identifier"),
                    syslog_pid=int_or_none(entry.get("syslog_pid")),
                    syslog_raw=entry.get("syslog_raw"),
                    documentation=entry.get("documentation"),
                    tid=int_or_none(entry.get("tid")),
                    unit=entry.get("unit"),
                    user_unit=entry.get("user_unit"),
                    pid=int_or_none(entry.get("pid")),
                    uid=int_or_none(entry.get("uid")),
                    gid=int_or_none(entry.get("gid")),
                    comm=entry.get("comm"),
                    exe=get_optional(entry.get("exe"), path_function),
                    cmdline=entry.get("cmdline"),
                    cap_effective=entry.get("cap_effective"),
                    audit_session=int_or_none(entry.get("audit_session")),
                    audit_loginuid=int_or_none(entry.get("audit_loginuid")),
                    systemd_cgroup=get_optional(entry.get("systemd_cgroup"), path_function),
                    systemd_slice=entry.get("systemd_slice"),
                    systemd_unit=entry.get("systemd_unit"),
                    systemd_user_unit=entry.get("systemd_user_unit"),
                    systemd_user_slice=entry.get("systemd_user_slice"),
                    systemd_session=entry.get("systemd_session"),
                    systemd_owner_uid=entry.get("systemd_owner_uid"),
                    selinux_context=entry.get("selinux_context"),
                    kernel_boot_id=entry.get("boot_id"),
                    machine_id=entry.get("machine_id"),
                    systemd_invocation_id=entry.get("systemd_invocation_id"),
                    transport=entry.get("transport"),
                    stream_id=entry.get("stream_id"),
                    line_break=entry.get("line_break"),
                    namespace=entry.get("namespace"),
                    runtime_scope=entry.get("runtime_scope"),
                    kernel_device=entry.get("kernel_device"),
                    kernel_subsystem=entry.get("kernel_subsystem"),
                    udev_sysname=entry.get("udev_sysname"),
                    udev_devnode=get_optional(entry.get("udev_devnode"), path_function),
                    udev_devlink=get_optional(entry.get("udev_devlink"), path_function),
                    journal_hostname=entry.get("hostname"),
                    source=journal_file,
                    _target=self.target,
                )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import DynamicDescriptor, TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.ssh.openssh import find_sshd_directory
from dissect.target.plugins.apps.ssh.ssh import SSHPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.plugins.general.config import ConfigurationTreePlugin
    from dissect.target.target import Target

SSHD_BOOLEAN_VALUES = (
    "yes",
    "no",
)

SSHD_BOOLEAN_FIELDS = (
    "AllowAgentForwarding",
    "DebianBanner",
    "DisableForwarding",
    "ChallengeResponseAuthentication",
    "ExposeAuthInfo",
    "GSSAPIAuthentication",
    "GSSAPICleanupCredentials",
    "GSSAPICleanupCredentials",
    "GSSAPIStrictAcceptorCheck",
    "GSSAPIStoreCredentialsOnRekey",
    "GSSAPIKeyExchange",
    "GSSAPIStrictAcceptorCheck",
    "GSSAPIStoreCredentialsOnRekey",
    "HostbasedAuthentication",
    "HostbasedUsesNameFromPacketOnly",
    "IgnoreUserKnownHosts",
    "KbdInteractiveAuthentication",
    "KeepAlive",
    "KerberosAuthentication",
    "KerberosGetAFSToken",
    "KerberosOrLocalPasswd",
    "KerberosTicketCleanup",
    "PasswordAuthentication",
    "PermitEmptyPasswords",
    "PermitTTY",
    "PermitUserRC",
    "PrintLastLog",
    "PrintMotd",
    "PubkeyAuthentication",
    "StreamLocalBindUnlink",
    "StrictModes",
    "TCPKeepAlive",
    "UsePAM",
    "X11Forwarding",
    "X11UseLocalhost",
)

SSHD_INTEGER_FIELDS = (
    "ClientAliveCountMax",
    "ClientAliveInterval",
    "LoginGraceTime",
    "MaxAuthTries",
    "MaxSessions",
    "Port",
    "X11DisplayOffset",
)

SSHD_MULTIPLE_DEFINITIONS_ALLOWED_FIELDS = (
    "AcceptEnv",
    "Include",
    "ListenAddress",
    "PermitListen",
    "PermitOpen",
    "Port",
)


class SSHServerPlugin(SSHPlugin):
    """OpenSSHd server plugin."""

    __namespace__ = "opensshd"

    def __init__(self, target: Target):
        super().__init__(target)
        self.sshd_directory = find_sshd_directory(target)
        self.sshd_config_path = self.sshd_directory.joinpath("sshd_config")

    def check_compatible(self) -> None:
        if not self.sshd_config_path.exists():
            raise UnsupportedPluginError("No sshd config found")

    @export(record=DynamicDescriptor(["datetime", "path"]))
    def config(self) -> Iterator[DynamicDescriptor]:
        """Parse all fields in the SSH server config in /etc/ssh/sshd_config.

        This function parses each line (not starting with '#') as a key-value
        pair, delimited by whitespace. The values of these fields can be one
        of three types: string, integer and boolean (string is the default).

        We provide two lists that define the integer and boolean fields
        (SSHD_INTEGER_FIELDS and SSHD_BOOLEAN_FIELDS).

        The fields in SSHD_MULTIPLE_DEFINITIONS_ALLOWED_FIELDS can be
        defined multiple times. We set their type to a list of the
        underlying value (e.g. varint[] for the Port field).

        This parser does not (yet) follow Include directives.

        References:
            - https://github.com/openssh/openssh-portable
            - https://www.man7.org/linux/man-pages/man5/sshd_config.5.html
        """

        record_fields = [
            ("datetime", "mtime"),
            ("path", "source"),
        ]

        config_tree: ConfigurationTreePlugin = self.target.config_tree

        # Parse sshd_config
        sshd_config = config_tree(
            path=self.sshd_config_path,
            collapse=SSHD_MULTIPLE_DEFINITIONS_ALLOWED_FIELDS,
            collapse_inverse=True,
            as_dict=True,
        )

        if _ := sshd_config.get("Include"):
            # Gather included config
            pass

        config = {}
        for key, value in sshd_config.items():
            if isinstance(value, dict):
                # A match statement, ignore for now
                continue
            _type, _value = _determine_type_and_value(key, value)

            config[key] = _value
            record_fields.append((_type, key))

        yield TargetRecordDescriptor("application/openssh/sshd_config", record_fields)(
            mtime=self.sshd_config_path.stat().st_mtime, source=self.sshd_config_path, **config, _target=self.target
        )


def _determine_type_and_value(key: str, value: str) -> tuple[str, Any]:
    _type = "string"
    _value = value

    _unpack = None
    if key in SSHD_INTEGER_FIELDS:
        _type = "varint"
        _unpack = int
    elif key in SSHD_BOOLEAN_FIELDS and _value.lower() in SSHD_BOOLEAN_VALUES:
        _type = "boolean"
        _unpack = _convert_bool
    else:
        _type = "string"

    if multiple_fields := (key in SSHD_MULTIPLE_DEFINITIONS_ALLOWED_FIELDS):
        _value = _value if isinstance(_value, list) else [_value]

    try:
        _value = _convert_function(_value, _unpack)
    except Exception:
        # Something went wrong, restore default type
        _type = "string"

    if multiple_fields:
        # The type can still be a list, so in that case, we append `[]`
        _type = f"{_type}[]"

    return _type, _value


def _convert_function(value: str | list, unpack_function: Callable | None) -> list[Any] | Any:
    if isinstance(value, list):
        return [unpack_function(val) for val in value]
    return unpack_function(value)


def _convert_bool(value: str) -> bool:
    return value.lower() == SSHD_BOOLEAN_VALUES[0]



from __future__ import annotations

import contextlib
import logging
from typing import TYPE_CHECKING, Callable
from unittest.mock import MagicMock, patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.vmx import VmxLoader
from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path


@contextlib.contextmanager
def _mock_vmx_and_container_open(disks: list[str]) -> Iterator[MagicMock]:
    with (
        patch("dissect.hypervisor.descriptor.vmx.VMX") as mock_vmx,
        patch("dissect.target.container.open") as mock_container_open,
    ):
        mock_vmx.parse.return_value = mock_vmx
        mock_vmx.disks.return_value = disks
        yield mock_container_open


@pytest.fixture
def mock_vmwarevm_dir(tmp_path: Path) -> Path:
    vm_path = tmp_path / "Test.vmwarevm"
    vm_path.mkdir()
    vmx_path = vm_path / "Test.vmx"
    vmx_path.touch()

    return vm_path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], mock_vmwarevm_dir: Path) -> None:
    """Test that we correctly use ``VmwarevmLoader`` when opening a ``Target``."""
    path = mock_vmwarevm_dir / "Test.vmx"

    with patch("dissect.target.container.open"), patch("dissect.target.target.Target.apply"):
        target = opener(path)
        assert isinstance(target._loader, VmxLoader)
        assert target.path == path


def test_loader(mock_vmwarevm_dir: Path) -> None:
    (mock_vmwarevm_dir / "mock.vmdk").touch()

    with _mock_vmx_and_container_open(["mock.vmdk"]) as mock_container_open:
        loader = loader_open(mock_vmwarevm_dir / "Test.vmx")
        assert isinstance(loader, VmxLoader)

        t = Target()
        loader.map(t)

        assert len(t.disks) == 1
        mock_container_open.assert_called_with(mock_vmwarevm_dir / "mock.vmdk")


def test_missing_disk(mock_vmwarevm_dir: Path, caplog: pytest.LogCaptureFixture) -> None:
    t = Target()

    with _mock_vmx_and_container_open(["mock.vmdk"]), caplog.at_level(logging.DEBUG, t.log.name):
        loader = loader_open(mock_vmwarevm_dir / "Test.vmx")
        assert isinstance(loader, VmxLoader)

        loader.map(t)
        assert len(t.disks) == 0
        assert "Disk not found: mock.vmdk" in caplog.text


def test_missing_snapshots(mock_vmwarevm_dir: Path, caplog: pytest.LogCaptureFixture) -> None:
    (mock_vmwarevm_dir / "mock.vmdk").touch()
    (mock_vmwarevm_dir / "mock-000001.vmdk").touch()

    t = Target()

    with (
        _mock_vmx_and_container_open(["mock-000002.vmdk"]) as mock_container_open,
        caplog.at_level(logging.DEBUG, t.log.name),
    ):
        loader = loader_open(mock_vmwarevm_dir / "Test.vmx")
        assert isinstance(loader, VmxLoader)

        loader.map(t)

        assert len(t.disks) == 1
        mock_container_open.assert_called_with(mock_vmwarevm_dir / "mock-000001.vmdk")

        assert "Disk not found but seems to be a snapshot, trying previous snapshots: mock-000002.vmdk" in caplog.text
        assert "Trying to load snapshot: mock-000001.vmdk" in caplog.text
        assert (
            "Missing disk(s) but continuing with found snapshot: mock-000001.vmdk (missing mock-000002.vmdk)"
            in caplog.text
        )


def test_missing_snapshots_base(mock_vmwarevm_dir: Path, caplog: pytest.LogCaptureFixture) -> None:
    (mock_vmwarevm_dir / "mock.vmdk").touch()

    t = Target()

    with (
        _mock_vmx_and_container_open(["mock-000002.vmdk"]) as mock_container_open,
        caplog.at_level(logging.DEBUG, t.log.name),
    ):
        loader = loader_open(mock_vmwarevm_dir / "Test.vmx")
        assert isinstance(loader, VmxLoader)

        loader.map(t)

        assert len(t.disks) == 1
        mock_container_open.assert_called_with(mock_vmwarevm_dir / "mock.vmdk")

        assert "Disk not found but seems to be a snapshot, trying previous snapshots: mock-000002.vmdk" in caplog.text
        assert "Trying to load snapshot: mock-000001.vmdk" in caplog.text
        assert "Trying to load snapshot: mock.vmdk" in caplog.text
        assert (
            "Missing disk(s) but continuing with found snapshot: mock.vmdk (missing mock-000002.vmdk, mock-000001.vmdk)"
            in caplog.text
        )


def test_missing_all_snapshots(mock_vmwarevm_dir: Path, caplog: pytest.LogCaptureFixture) -> None:
    t = Target()

    with _mock_vmx_and_container_open(["mock-000001.vmdk"]), caplog.at_level(logging.DEBUG, t.log.name):
        loader = loader_open(mock_vmwarevm_dir / "Test.vmx")
        assert isinstance(loader, VmxLoader)

        loader.map(t)

        assert len(t.disks) == 0

        assert "Failed to find previous snapshot for disk: mock-000001.vmdk" in caplog.text



from io import BytesIO

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugin import OperatingSystem
from dissect.target.plugins.os.unix.linux.redhat._os import RedHatPlugin
from dissect.target.target import Target


@pytest.mark.parametrize(
    "file_name",
    [
        ("/etc/redhat-release"),
        ("/etc/centos-release"),
        ("/etc/fedora-release"),
        ("/etc/sysconfig/network-scripts"),
    ],
)
def test_unix_linux_redhat_os_detection(target_bare: Target, file_name: str) -> None:
    """Test if we detect RedHat OS correctly."""

    fs = VirtualFilesystem()
    fs.map_file_fh(file_name, BytesIO())

    target_bare.filesystems.add(fs)
    target_bare.apply()

    assert RedHatPlugin.detect(target_bare)
    assert isinstance(target_bare._os, RedHatPlugin)
    assert target_bare.os == OperatingSystem.LINUX



from __future__ import annotations

import logging
import os
import traceback
import urllib.parse
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar, Union

from dissect.target import container, filesystem, loader, plugin, volume
from dissect.target.exceptions import (
    FilesystemError,
    PluginError,
    PluginNotFoundError,
    TargetError,
    TargetPathNotFoundError,
    UnsupportedPluginError,
    VolumeSystemError,
)
from dissect.target.helpers import config
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.loaderutil import parse_path_uri
from dissect.target.helpers.utils import StrEnum, slugify
from dissect.target.loaders.direct import DirectLoader
from dissect.target.plugins.os.default._os import DefaultOSPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.helpers.record import ChildTargetRecord

log = logging.getLogger(__name__)

FunctionTuple = tuple[plugin.Plugin, Union[plugin.Plugin, property, None]]


class Event(StrEnum):
    INCOMPATIBLE_PLUGIN = "incompatible-plugin"
    REGISTERED_PLUGIN = "registered-plugin"
    FUNC_EXEC = "function-execution"
    FUNC_EXEC_ERROR = "function-execution-error"


def getlogger(target: Target) -> TargetLogAdapter:
    if not log.root.handlers:
        log.setLevel(os.getenv("DISSECT_LOG_TARGET", "CRITICAL"))
    return TargetLogAdapter(log, {"target": target})


class TargetLogAdapter(logging.LoggerAdapter):
    def process(self, msg: str, kwargs: dict[str, Any]) -> tuple[str, dict[str, Any]]:
        return f"{self.extra['target']}: {msg}", kwargs


class Target:
    """The class that represents the target that you are talking to.

    Usually you do not want to instantiate this class directly, but rather use the
    :meth:`Target.open` or :meth:`Target.open_all` methods to open a target from a path or URI.
    The target will then be loaded using the appropriate :class:`Loader <dissect.target.loader.Loader>`.

    Targets are the glue that connects the different ``Containers``, ``Loaders``, ``Volumes``
    and ``Filesystems`` together.
    ``Loaders`` are used to map the ``Containers``, ``Volumes`` and ``Filesystems`` of the target
    onto the ``Target`` object.

    The plugins of dissect.target get mapped onto the ``Target`` too.
    They become available as attributes on a ``Target`` object. For example, ``t.hostname``, ``t.evtx()``.
    By executing the plugin function with a target, it will perform the function on itself.

    Args:
        path: The path of a target.
    """

    def __init__(self, path: str | Path | None = None):
        self.path_scheme = None
        self.path_query = {}

        self.path, self.parsed_path = parse_path_uri(path)
        if self.parsed_path:
            # Treat everything after the @ in a URI as part of the path
            # I.e. "uri://user:password@host:port/path" becomes "host:port/path"
            self.path = Path(self.parsed_path.netloc.split("@", 1)[-1] + self.parsed_path.path)
            self.path_scheme = self.parsed_path.scheme
            self.path_query = dict(urllib.parse.parse_qsl(self.parsed_path.query, keep_blank_values=True))

        self.props: dict[Any, Any] = {}
        self.log = getlogger(self)

        self._name = None
        self._plugins: list[plugin.Plugin] = []
        self._functions: dict[str, FunctionTuple] = {}
        self._loader: loader.Loader = None
        self._os = None
        self._os_plugin: type[plugin.OSPlugin] = None
        self._child_plugins: dict[str, plugin.ChildTargetPlugin] = {}
        self._cache = {}
        self._errors = []
        self._applied = False

        # We do not want to look for config files at the Target path if it is actually a child Target.
        config_paths = [Path.cwd(), Path.home()]
        if not isinstance(path, TargetPath):
            config_paths = [self.path, *config_paths]

        try:
            self._config = config.load(config_paths)
        except Exception as e:
            self.log.warning("Error loading config file: %s", config_paths)
            self.log.debug("", exc_info=e)
            self._config = config.load(None)  # This loads an empty config.

        # Fill the disks and/or volumes and/or filesystems and apply() will
        # make sure that volumes/filesystems gets filled and the filesystems
        # get auto mounted on top of fs.
        #
        # Except when a volume's guid or volumesystem's serial is unknown (in
        # the case of windows), in which case _init_os() and
        # _os_plugin.create() won't be able to figure out where to mount the
        # filesystem. In that case they should be added and mounted explicitly
        # by the loader.
        #
        # A collection of instances of Container() subclasses
        self.disks = DiskCollection(self)
        # A collection of instances of Volume() subclasses
        self.volumes = VolumeCollection(self)
        # A collection of instances of Filesystem() subclasses
        self.filesystems = FilesystemCollection(self)

        self.fs = filesystem.RootFilesystem(self)

    def __repr__(self) -> str:
        return f"<Target {self.path}>"

    def __getattr__(self, attr: str) -> plugin.Plugin | Any:
        """Override of the default ``__getattr__`` so plugins and functions can be called from a ``Target`` object."""
        p, func = self.get_function(attr)

        if isinstance(func, property):
            # If it's a property, execute it and return the result
            try:
                result = func.__get__(p)
                self.send_event(Event.FUNC_EXEC, func=attr)
            except Exception:
                if not attr.startswith("__"):
                    self.send_event(
                        Event.FUNC_EXEC_ERROR,
                        func=attr,
                        stacktrace=traceback.format_exc(),
                    )
                raise
            else:
                return result

        return func

    def __dir__(self) -> list[str]:
        """Override the default ``__dir__`` to provide autocomplete for things like IPython."""
        funcs = []
        if self._os_plugin:
            funcs = list(self._os_plugin.__functions__)

        for plugin_desc in plugin.plugins(self._os_plugin):
            funcs.extend(plugin_desc.functions)

        result = set(self.__dict__.keys())
        result.update(self.__class__.__dict__.keys())
        result.update(object.__dict__.keys())
        result.update(funcs)

        return list(result)

    @classmethod
    def set_event_callback(cls, *, event_type: Event | None = None, event_callback: Callable) -> None:
        """Sets ``event_callbacks`` on a Target class.

        ``event_callbacks`` get used to handle specific events denoted by :class:`Event`.
        This records events related to the target, such as:

        - a plugin gets registered to the target
        - a plugin is incompatible with the target
        - a function succeededs in its execution
        - a function fails in execution
        """
        if not hasattr(cls, "event_callbacks"):
            cls.event_callbacks = defaultdict(set)

        # When `event_type` is not set or is None, the callback is a catch-all callback
        cls.event_callbacks[event_type].add(event_callback)

    def send_event(self, event_type: Event, **kwargs) -> None:
        """Notify event callbacks for the given ``event_type``.

        Each event can have multiple callback methods, it calls all the callbacks that fit the corresponding event type.
        ``None`` is a catch-all method for event callbacks that always get called.

        Args:
            event_type: The type of event
        """
        cls = type(self)
        if not hasattr(cls, "event_callbacks"):
            return

        specific_callbacks = cls.event_callbacks.get(event_type, set())
        catch_all_callbacks = cls.event_callbacks.get(None, set())

        for callback in set.union(specific_callbacks, catch_all_callbacks):
            try:
                callback(self, event_type, **kwargs)
            except Exception:  # noqa: PERF203
                self.log.warning("Can't send event %s to %s", event_type, callback, exc_info=True)

    def apply(self) -> None:
        """Resolve all disks, volumes and filesystems and load an operating system on the current ``Target``."""
        self.disks.apply()
        self.volumes.apply()
        self.filesystems.apply()
        self._init_os()
        self._mount_others()
        self._applied = True

    @property
    def _generic_name(self) -> str:
        """Return a generic name for this target."""

        generic_name = (self.path and self.path.name) or None

        if not generic_name:
            # The os() plugin function is only available after a Target.apply(), as
            # only then an _os_plugin is assigned.
            try:
                generic_name = f"Unknown-{self.os}"
            except Exception:
                generic_name = "Unknown"

        return generic_name

    @property
    def name(self) -> str:
        """Return a name for this target.

        The function is guaranteed to give back some name.
        The name will be guaranteed to not have slashes, backslashes and spaces.
        The name won't be guaranteed to be unique.

        Returns:
            The name of a target.
        """
        if self._name:
            return self._name

        target_name = None
        try:
            target_name = self.hostname
        except Exception:
            pass
        finally:
            if not target_name:
                target_name = self._generic_name

        target_name = slugify(target_name)

        # As name() may be called before the Target instance is fully
        # instantiated (before apply() is called), we only save the name once
        # all parts of the Target instance are there, meaning an _os_plugin is
        # present, and the name won't change anymore.
        if self._applied:
            self._name = target_name

        return target_name

    @classmethod
    def open(cls, path: str | Path) -> Self:
        """Try to find a suitable loader for the given path and load a ``Target`` from it.

        Args:
            path: Path to load the ``Target`` from.
                  If the path is a ``os.PathLike`` object, it will be used as-is.
                  If the path is a string and looks like a URI, it will be parsed as such.
                  If the path is a string and does not like like a URI, it will be treated as a local path.

        Returns:
            A Target with a linked :class:`~dissect.target.loader.Loader` object.
        """

        spec = path

        # If the path is a URI-like string, separate the path component
        adjusted_path, parsed_path = parse_path_uri(spec)
        # We always need a path to work with, so convert the spec into one if it's not one already
        path = Path(spec) if not isinstance(spec, os.PathLike) else spec

        if parsed_path is not None and (loader_cls := loader.find_loader_by_scheme(parsed_path.scheme)):
            # If we find a loader by URI scheme, use the adjusted path (path component of the URI)
            found_path = adjusted_path
        elif loader_cls := loader.find_loader(path, fallbacks=[loader.DirLoader, loader.RawLoader]):
            # Otherwise try to find a loader for the "raw" path
            # If we succeed, upgrade the "spec" to the path
            spec = path
            found_path = path
            parsed_path = None
        else:
            # No loader found
            return cls.open_raw(spec)

        try:
            loader_instance = loader_cls(found_path, parsed_path=parsed_path)
        except Exception as e:
            raise TargetError(f"Failed to initiate {loader_cls.__name__} for target {spec}: {e}") from e

        return cls._load(spec, loader_instance)

    @classmethod
    def open_raw(cls, path: str | Path) -> Self:
        """Open a Target with the given path using the :class:`~dissect.target.loaders.raw.RawLoader`.

        Args:
            path: Path to load the ``Target`` from.
        """
        adjusted_path = Path(path) if not isinstance(path, os.PathLike) else path
        return cls._load(path, loader.RawLoader(adjusted_path))

    @classmethod
    def open_all(cls, paths: str | Path | list[str | Path], include_children: bool = False) -> Iterator[Self]:
        """Yield all targets from one or more paths or directories.

        If the path is a directory, iterate files one directory deep.

        Args:
            paths: A list of paths to load ``Targets`` from.
                   If the path is a ``os.PathLike`` object, it will be used as-is.
                   If the path is a string and looks like a URI, it will be parsed as such.
                   If the path is a string and does not look like a URI, it will be treated as a local path.
            include_children: Whether to recursively open child targets.

        Raises:
            TargetError: Raised when not a single ``Target`` can be loaded.
        """

        def _open_all(spec: str | Path, include_children: bool = False) -> Iterator[Target]:
            # If the path is a URI-like string, separate the path component
            adjusted_path, parsed_path = parse_path_uri(spec)
            # We always need a path to work with, so convert the spec into one if it's not one already
            path = Path(spec) if not isinstance(spec, os.PathLike) else spec

            if parsed_path is not None and (loader_cls := loader.find_loader_by_scheme(parsed_path.scheme)):
                # If we find a loader by URI scheme, use the adjusted path (path component of the URI)
                found_path = adjusted_path
            elif loader_cls := loader.find_loader(path, fallbacks=[loader.DirLoader, loader.RawLoader]):
                # Otherwise try to find a loader for the "raw" path
                # If we succeed, upgrade the "spec" to the path
                spec = path
                found_path = path
                parsed_path = None
            else:
                # Couldn't find a loader
                return

            getlogger(spec).debug("Attempting to use loader: %s", loader_cls)
            # See if the loader provides any additional paths to load
            # This is useful for URI-like paths that have wildcard components
            for sub_path in loader_cls.find_all(found_path, parsed_path=parsed_path):
                if sub_path is found_path:
                    # If the sub path is the same, the loader didn't provide any additional paths
                    # Use the original spec and paths
                    load_spec = spec
                    load_path = found_path
                    load_parsed_path = parsed_path
                else:
                    # Otherwise, use the sub path as the spec and optionally parse it as a URI
                    load_spec = sub_path
                    if isinstance(sub_path, str):
                        load_path, load_parsed_path = parse_path_uri(sub_path)
                    else:
                        load_path = sub_path
                        load_parsed_path = parsed_path

                try:
                    # Try to instantiate the loader
                    # For file/dir-like specs, load_path is that file/dir, and parsed_path is None
                    # For URI-like specs, load_path is the path component of the URI, and parsed_path is the parsed URI
                    ldr = loader_cls(load_path, parsed_path=load_parsed_path)
                except Exception as e:
                    message = "%s" if isinstance(e, TargetPathNotFoundError) else "Failed to initiate loader: %s"
                    getlogger(load_spec).error(message, e)
                    getlogger(load_spec).debug("", exc_info=e)
                    continue

                try:
                    # Attempt to load the target using this loader
                    # load_spec is the original "spec"
                    # For URI-like paths it's the full original URI
                    # For file/dir-like paths it's a Path object
                    target = cls._load(load_spec, ldr)
                except Exception as e:
                    getlogger(load_spec).error("Failed to load target with loader %s", ldr, exc_info=e)
                    continue
                else:
                    yield target

                if include_children:
                    try:
                        yield from target.open_children()
                    except Exception as e:
                        getlogger(load_spec).error("Failed to load child target from %s", target, exc_info=e)

        at_least_one_loaded = False

        if isinstance(paths, (str, Path)):
            paths = [paths]

        # Treat every path as a unique target spec
        for spec in paths:
            loaded = False

            for target in _open_all(spec, include_children=include_children):
                loaded = True
                at_least_one_loaded = True
                yield target

            if loaded:
                # If we found a loader for the top level path, no need to search a level deeper
                # Going deeper could cause unwanted behaviour
                continue

            # If we did not find a loader for the top level path, we will search one directory deep
            # This only works for directories and non URI-like paths
            _, parsed_path = parse_path_uri(spec)
            if parsed_path is None or isinstance(spec, os.PathLike):
                path = Path(spec) if not isinstance(spec, os.PathLike) else spec

                for entry in path.iterdir():
                    for target in _open_all(entry, include_children=include_children):
                        at_least_one_loaded = True
                        yield target

        if not at_least_one_loaded:
            raise TargetError(f"Failed to find any loader for targets: {paths}")

    @classmethod
    def open_direct(cls, paths: list[str | Path]) -> Self:
        """Create a minimal target with a virtual root filesystem with all ``paths`` mapped into it.

        This is useful when running plugins on individual files.
        """
        return cls._load("direct", DirectLoader(paths))

    @property
    def is_direct(self) -> bool:
        """Check if the target is a direct target."""
        return isinstance(self._loader, DirectLoader)

    def _load_child_plugins(self) -> None:
        """Load special :class:`~dissect.target.plugin.ChildTargetPlugin` plugins.

        These plugins inform the ``Target`` how to deal with child targets, such as VMs from a hypervisor.
        Examples of these plugins are:

        - :class:`~dissect.target.plugins.child.esxi.ESXiChildTargetPlugin`
        - :class:`~dissect.target.plugins.child.hyper-v.HyperVChildTargetPlugin`
        """
        if self._child_plugins:
            return

        for plugin_desc in plugin.child_plugins():
            try:
                plugin_cls: type[plugin.ChildTargetPlugin] = plugin.load(plugin_desc)
                child_plugin = plugin_cls(self)
            except PluginError:
                self.log.exception("Failed to load child plugin: %s", plugin_desc.qualname)
                continue
            except Exception:
                self.log.exception("Broken child plugin: %s", plugin_desc.qualname)
                continue

            try:
                child_plugin.check_compatible()
                self._child_plugins[child_plugin.__type__] = child_plugin
            except PluginError as e:
                self.log.info("Child plugin reported itself as incompatible: %s", plugin_desc.qualname)
                self.log.debug("", exc_info=e)
            except Exception:
                self.log.exception(
                    "An exception occurred while checking for child plugin compatibility: %s", plugin_desc.qualname
                )

    def open_child(self, child: str | Path) -> Target:
        """Open a child target.

        Args:
            child: The location of a target within the current ``Target``.

        Returns:
            An opened ``Target`` object of the child target.
        """
        if isinstance(child, str) and child.isdecimal():
            child_num = int(child)
            for child_record in self.list_children():
                if child_num == 0:
                    return Target.open(self.fs.path(child_record.path))
                child_num -= 1
            raise IndexError("Child target index out of range")

        return Target.open(self.fs.path(child))

    def open_children(self, recursive: bool = False) -> Iterator[Target]:
        """Open all the child targets on a ``Target``.

        Will open all discovered child targets if the current ``Target`` has them, such as VMs on a hypervisor.

        Args:
            recursive: Whether to check the child ``Target`` for more ``Targets``.

        Returns:
            An iterator of ``Targets``.
        """
        for child in self.list_children():
            try:
                target = self.open_child(child.path)
            except TargetError:
                self.log.exception("Failed to open child target %s", child)
                continue

            yield target

            if recursive:
                yield from target.open_children(recursive=recursive)

    def list_children(self) -> Iterator[ChildTargetRecord]:
        """Lists all child targets that compatible :class:`~dissect.target.plugin.ChildTargetPlugin` classes
        can discover.
        """
        self._load_child_plugins()
        for child_plugin in self._child_plugins.values():
            yield from child_plugin.list_children()

    def reload(self) -> Target:
        """Reload the current target.

        Using the loader with which the target was originally loaded,
        reload the path of the current target.
        This is useful when the target is live.

        Raises:
            TargetError: If the target has no path or loader.

        Returns:
            A fresh ``Target`` object
        """

        if self._loader and self.path:
            return self._load(self.path, self._loader)

        raise TargetError("Target has no path and/or loader")

    @classmethod
    def _load(cls, path: str | Path | None, ldr: loader.Loader) -> Self:
        """Internal function that attemps to load a path using a given loader.

        Args:
            path: The path to the target.
            ldr: The loader to use for loading this target.

        Raises:
            TargetError: If it failed to load a target.

        Returns:
            A ``Target`` object with disks, volumes and/or filesystems mapped by the ``ldr`` from the given ``path``.
        """
        target = cls(path)

        try:
            ldr.map(target)
            target._loader = ldr
            target.apply()
        except Exception as e:
            raise TargetError(f"Failed to load target: {path}") from e
        else:
            return target

    def _init_os(self) -> None:
        """Internal function that attemps to load an OSPlugin for this target."""
        if self._os_plugin:
            # If self._os_plugin is already assigned, we expect it to be fully
            # configured and possibly already instantiated (if not, it will
            # be), hence no os_plugin.create() is run and no detection is
            # attempted.
            os_plugin = self._os_plugin

            if isinstance(os_plugin, plugin.OSPlugin):
                self._os_plugin = os_plugin.__class__
            elif issubclass(os_plugin, plugin.OSPlugin) and (fs := os_plugin.detect(self)):
                os_plugin = os_plugin.create(self, fs)

            self._os = self.add_plugin(os_plugin)
            return

        if not len(self.disks) and not len(self.volumes) and not len(self.filesystems):
            raise TargetError(f"Failed to load target. No disks, volumes or filesystems: {self.path}")

        candidates: list[tuple[plugin.PluginDescriptor, type[plugin.OSPlugin], filesystem.Filesystem]] = []

        for plugin_desc in plugin.os_plugins():
            # Subclassed OS Plugins used to also subclass the detection of the
            # parent. This meant that in order for a subclassed OS Plugin to be a
            # candidate for the final OS Plugin, the parent OS Plugin's detection
            # had to succeed. This caused some complications in cases where the
            # filesystem layout needed some tweaks by a subclassed OS Plugin before
            # parent OS Plugin paths made sense.
            #
            # E.g. VyOS /boot/<version>/live-rw -> map to /
            #
            # Now subclassed OS Plugins are on the same detection "layer" as
            # regular OS Plugins, but can still inherit functions.
            qualname = plugin_desc.qualname
            self.log.debug("Loading OS plugin: %s", qualname)
            try:
                os_plugin: type[plugin.OSPlugin] = plugin.load(plugin_desc)
                fs = os_plugin.detect(self)
            except PluginError:
                self.log.exception("Failed to load OS plugin: %s", qualname)
                continue
            except Exception:
                self.log.exception("Broken OS plugin: %s", qualname)
                continue

            if not fs:
                continue

            self.log.info("Found compatible OS plugin: %s", qualname)
            candidates.append((plugin_desc, os_plugin, fs))

        fs = None
        os_plugin = DefaultOSPlugin

        if candidates:
            plugin_desc, os_plugin, fs = candidates[0]
            for candidate_plugin_desc, candidate_plugin, candidate_fs in candidates[1:]:
                # More specific OS plugins are considered better candidates
                if len(candidate_plugin.mro()) > len(os_plugin.mro()):
                    plugin_desc, os_plugin, fs = candidate_plugin_desc, candidate_plugin, candidate_fs

            self.log.debug("Selected OS plugin: %s", plugin_desc.qualname)
        else:
            # No OS detected
            self.log.warning("Failed to find OS plugin, falling back to default")

        self._os_plugin = os_plugin
        self._os = self.add_plugin(os_plugin.create(self, fs))

    def _mount_others(self) -> None:
        root_fs = self.fs
        counter = 0
        path = "/$fs$/fs0"

        for fs in self.filesystems:
            if fs not in root_fs.mounts.values():
                # determine mount point
                while root_fs.path(path).exists():
                    counter += 1
                    path = f"/$fs$/fs{counter}"

                root_fs.mount(path, fs)

    def add_plugin(
        self,
        plugin_cls: plugin.Plugin | type[plugin.Plugin],
        check_compatible: bool = True,
    ) -> plugin.Plugin:
        """Add and register a plugin by class.

        Args:
            plugin_cls: The plugin to add and register, this can either be a class or instance. When this is a class,
                        it will be instantiated.
            check_compatible: Check whether the plugin is compatible with the ``Target``.

        Returns:
            The ``plugin_cls`` instance.

        Raises:
            UnsupportedPluginError: Raised when plugins were found, but they were incompatible
            PluginError: Raised when any other exception occurs while trying to load the plugin.
        """
        self.log.debug("Adding plugin: %s", plugin_cls)

        if not isinstance(plugin_cls, plugin.Plugin):
            try:
                p = plugin_cls(self)
            except PluginError:
                raise
            except Exception as e:
                raise PluginError(f"An exception occurred while trying to initialize a plugin: {plugin_cls}") from e
        else:
            p = plugin_cls

        if not isinstance(p, plugin.Plugin):
            raise PluginError(f"Not a subclass of Plugin: {p}")

        if check_compatible and not self.is_direct:
            try:
                p.check_compatible()
            except PluginError:
                self.send_event(Event.INCOMPATIBLE_PLUGIN, plugin_cls=plugin_cls)
                raise
            except Exception as e:
                raise UnsupportedPluginError(
                    f"An exception occurred while checking for plugin compatibility: {plugin_cls}"
                ) from e

        self._register_plugin_functions(p)

        return p

    def load_plugin(self, descriptor: plugin.PluginDescriptor | plugin.FunctionDescriptor) -> plugin.Plugin:
        """Load a plugin by descriptor.

        Args:
            plugin_desc: The descriptor of the plugin to load.

        Returns:
            The loaded plugin instance.

        Raises:
            PluginError: Raised when any exception occurs while trying to load the plugin.
        """
        return self.add_plugin(plugin.load(descriptor))

    def _register_plugin_functions(self, plugin_inst: plugin.Plugin) -> None:
        """Internal function that registers all the exported functions from a given plugin.

        Args:
            plugin_inst: Instance of a plugin.
        """
        self.send_event(Event.REGISTERED_PLUGIN, plugin_inst=plugin_inst)

        self._plugins.append(plugin_inst)

        if plugin_inst.__namespace__:
            self._functions[plugin_inst.__namespace__] = (plugin_inst, plugin_inst)

            for func in plugin_inst.__functions__:
                self._functions[f"{plugin_inst.__namespace__}.{func}"] = (plugin_inst, None)
        else:
            for func in plugin_inst.__functions__:
                # If we getattr here, property members will be executed, so we do that in __getattr__
                self._functions[func] = (plugin_inst, None)

    def get_function(self, function: str | plugin.FunctionDescriptor) -> FunctionTuple:
        """Attempt to get a given function.

        If the function is not already registered, look for plugins that export the function and register them.

        Args:
            function: Function name to look for.

        Returns:
            A tuple of the plugin and the corresponding function.

        Raises:
            UnsupportedPluginError: Raised when plugins were found, but they were incompatible
            PluginError: Raised when any other exception occurs while trying to load the plugin.
        """
        if isinstance(function, plugin.FunctionDescriptor):
            function_name = function.name

            if function_name not in self._functions:
                try:
                    self.load_plugin(function)
                except UnsupportedPluginError:
                    self.send_event(Event.INCOMPATIBLE_PLUGIN, plugin_desc=function)
                    raise UnsupportedPluginError(
                        f"Unsupported function `{function.name}` (`{function.module}`)",
                    )

        else:
            function_name = function

            if function not in self._functions:
                causes = []

                descriptor = None
                for descriptor in plugin.lookup(function, self._os_plugin):
                    try:
                        self.load_plugin(descriptor)
                        self.log.debug("Found compatible plugin '%s' for function '%s'", descriptor.qualname, function)
                        break
                    except UnsupportedPluginError as e:
                        self.send_event(Event.INCOMPATIBLE_PLUGIN, plugin_desc=descriptor)
                        causes.append(e)
                else:
                    if descriptor:
                        # In this case we made at least one iteration but it was skipped due incompatibility.
                        # Just take the last known cause for now
                        raise UnsupportedPluginError(
                            f"Unsupported function `{function}` for target with OS plugin {self._os_plugin}",
                            extra=causes[1:] if len(causes) > 1 else None,
                        ) from (causes[0] if causes else None)

        # We still ended up with no compatible plugins
        if function_name not in self._functions:
            raise PluginNotFoundError(f"Can't find plugin with function `{function_name}`")

        p, func = self._functions[function_name]
        if func is None:
            method_attr = function_name.rpartition(".")[2] if p.__namespace__ else function_name
            func = getattr(p.__class__, method_attr)

            if not isinstance(func, property) or (
                isinstance(func, property) and getattr(func.fget, "__persist__", False)
            ):
                # If the persist flag is set on a property, store the property result in the function cache
                # This is so we don't have to evaluate the property again
                func = getattr(p, method_attr)
            self._functions[function_name] = (p, func)

        return p, func

    def has_function(self, function: str) -> bool:
        """Return whether this Target supports a given function.

        Args:
            function: The function name to look for.

        Returns:
            ``True`` if the function can be found, ``False`` otherwise.
        """
        try:
            self.get_function(function)
        except PluginError:
            return False
        else:
            return True


T = TypeVar("T")


class Collection(Generic[T]):
    def __init__(self, target: Target):
        self.target = target
        self.entries: list[T] = []

    def add(self, entry: T) -> None:
        self.entries.append(entry)

    def __getitem__(self, k: str) -> T:
        return self.entries[k]

    def __iter__(self) -> Iterator[T]:
        return iter(self.entries)

    def __len__(self) -> int:
        return len(self.entries)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.entries!r}>"


class DiskCollection(Collection[container.Container]):
    def apply(self) -> None:
        for disk in self.entries:
            # Some LVM configurations (i.e. RAID with the metadata at the end of the disk)
            # may be misidentified as having a valid MBR/GPT on some of the disks
            # To counter this, first check if the disk is part of any LVM configurations that we support
            if not volume.is_lvm_volume(disk):
                try:
                    if not hasattr(disk, "vs") or disk.vs is None:
                        disk.vs = volume.open(disk)
                        self.target.log.debug("Opened volume system: %s on %s", disk.vs, disk)

                    if not len(disk.vs.volumes):
                        raise VolumeSystemError("Volume system has no volumes")  # noqa: TRY301

                    for vol in disk.vs.volumes:
                        self.target.volumes.add(vol)
                    continue
                except Exception as e:
                    self.target.log.warning(
                        "Can't identify volume system or no volumes found, adding as raw volume instead: %s", disk
                    )
                    self.target.log.debug("", exc_info=e)

            # Fallthrough case for error and if we're part of a logical volume set
            vol = volume.Volume(disk, 1, 0, disk.size, None, None, disk=disk)
            self.target.volumes.add(vol)


class VolumeCollection(Collection[volume.Volume]):
    def apply(self) -> None:
        # We don't want later additions to modify the todo, so make a copy
        todo = self.entries[:]
        fs_volumes = []
        lvm_volumes = []
        encrypted_volumes = []

        while todo:
            new_volumes = []
            lvm_volumes = []
            encrypted_volumes = []

            for vol in todo:
                if volume.is_lvm_volume(vol):
                    lvm_volumes.append(vol)
                elif volume.is_encrypted(vol):
                    encrypted_volumes.append(vol)
                else:
                    # We could be getting "regular" volume systems out of LVM or encrypted volumes
                    # Try to open each volume as a regular volume system, or add as a filesystem if it fails
                    # There are a few scenarios were we want to discard the opened volume, though
                    #
                    # If the current volume offset is 0 and originates from a "regular" volume system, we're likely
                    # opening a volume system on the same disk again
                    # Sometimes BSD systems are configured this way and an FFS volume "starts" at offset 0
                    #
                    # If we opened an empty volume system, it might also be the case that a filesystem actually
                    # "starts" at offset 0

                    # Regardless of what happens, we want to try to open it as a filesystem later on
                    fs_volumes.append(vol)

                    if vol.offset == 0 and vol.vs and vol.vs.__type__ == "disk":
                        # We are going to re-open a volume system on itself, bail out
                        self.target.log.info("Found volume with offset 0, opening as raw volume instead")
                        continue

                    try:
                        vs = volume.open(vol)
                    except Exception:
                        # If opening a volume system fails, there's likely none, so open as a filesystem instead
                        continue

                    if not len(vs.volumes):
                        # We opened an empty volume system, discard
                        continue

                    self.entries.extend(vs.volumes)
                    new_volumes.extend(vs.volumes)

            self.target.log.debug("LVM volumes found: %s", lvm_volumes)
            self.target.log.debug("Encrypted volumes found: %s", encrypted_volumes)

            for lvm in volume.open_lvm(lvm_volumes):
                self.target.log.debug("Opened LVM: %s", lvm)
                for lv in lvm.volumes:
                    self.add(lv)
                    new_volumes.append(lv)

            for enc_volume in encrypted_volumes:
                for dec_volume in volume.open_encrypted(enc_volume):
                    self.target.log.debug("Encrypted volume opened: %s", enc_volume)
                    self.add(dec_volume)
                    new_volumes.append(dec_volume)

            todo = new_volumes

        mv_fs_volumes = []
        for vol in fs_volumes:
            try:
                if getattr(vol, "fs", None) is None:
                    if filesystem.is_multi_volume_filesystem(vol):
                        mv_fs_volumes.append(vol)
                    else:
                        vol.fs = filesystem.open(vol)
                        self.target.log.debug("Opened filesystem: %s on %s", vol.fs, vol)

                if getattr(vol, "fs", None) is not None:
                    self.target.filesystems.add(vol.fs)
            except FilesystemError as e:  # noqa: PERF203
                vol.seek(0)
                if not vol.read(8 * 1024 * 1024).strip(b"\x00"):
                    # If the first 8MB is empty, don't log a warning
                    self.target.log.debug("Skipping empty volume: %s", vol)
                else:
                    self.target.log.warning("Can't identify filesystem: %s", vol)
                    self.target.log.debug("", exc_info=e)

        for fs in filesystem.open_multi_volume(mv_fs_volumes):
            self.target.filesystems.add(fs)
            for vol in fs.volume:
                vol.fs = fs


class FilesystemCollection(Collection[filesystem.Filesystem]):
    def apply(self) -> None:
        for fs in self.entries:
            for subfs in fs.iter_subfs():
                self.add(subfs)



from __future__ import annotations

import zlib
from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

# The structs rawheader and rawrecord have added padding compared to the structs in the source code
# https://github.com/Atoptool/atop/blob/master/rawlog.h, https://github.com/Atoptool/atop/blob/master/photoproc.h
atop_def = """
typedef unsigned long long time_t;
typedef long long count_t;

#define  _UTSNAME_LENGTH  65

struct utsname {
    char  sysname[_UTSNAME_LENGTH];     /* Name of the implementation of the operating system. */
    char  nodename[_UTSNAME_LENGTH];    /* Name of this node on the network. */
    char  release[_UTSNAME_LENGTH];     /* Current release level of this implementation. */
    char  version[_UTSNAME_LENGTH];     /* Current version level of this release. */
    char  machine[_UTSNAME_LENGTH];     /* Name of the hardware type the system is running on. */
    char  domainname[_UTSNAME_LENGTH];  /* Name of the domain of this node on the network. */
};

struct rawheader {
    unsigned  int      magic;
    unsigned  short    aversion;        /* creator atop version with MSB */
    unsigned  short    future1;         /* can be reused */
    unsigned  short    future2;         /* can be reused */
    unsigned  short    rawheadlen;      /* length of struct rawheader */
    unsigned  short    rawreclen;       /* length of struct rawrecord */
    unsigned  short    hertz;           /* clock interrupts per second */
    unsigned  short    sfuture[6];      /* future use */
    unsigned  int      sstatlen;        /* length of struct sstat */
    unsigned  int      tstatlen;        /* length of struct tstat */
    struct    utsname  utsname;         /* info about this system  */
    char               cfuture[8];      /* future use */
    unsigned  int      pagesize;        /* size of memory page (bytes) */
    int                supportflags;    /* used features */
    int                osrel;           /* OS release number */
    int                osvers;          /* OS version number */
    int                ossub;           /* OS version subnumber */
    int                ifuture[6];      /* future use */
    uint16             padding;
};

struct rawrecord {
    time_t           curtime;           /* current time (epoch) */
    unsigned  short  flags;             /* various flags */
    unsigned  short  sfuture[3];        /* future use */
    unsigned  int    scomplen;          /* length of compressed sstat */
    unsigned  int    pcomplen;          /* length of compressed tstat's */
    unsigned  int    interval;          /* interval (number of seconds) */
    unsigned  int    ndeviat;           /* number of tasks in list */
    unsigned  int    nactproc;          /* number of processes in list */
    unsigned  int    ntask;             /* total number of tasks */
    unsigned  int    totproc;           /* total number of processes */
    unsigned  int    totrun;            /* number of running  threads */
    unsigned  int    totslpi;           /* number of sleeping threads(S) */
    unsigned  int    totslpu;           /* number of sleeping threads(D) */
    unsigned  int    totzomb;           /* number of zombie processes */
    unsigned  int    nexit;             /* number of exited processes */
    unsigned  int    noverflow;         /* number of overflow processes */
    unsigned  int    ifuture[6];        /* future use */
    int              padding;
};
"""
atop_tstat_def = """
#define  PNAMLEN  15
#define  CMDLEN   255

/* structure containing only relevant process-info extracted from kernel's process-administration */
struct tstat {
    /* GENERAL TASK INFO */
    struct gen {
        int     tgid;                   /* threadgroup identification */
        int     pid;                    /* process identification */
        int     ppid;                   /* parent process identification */
        int     ruid;                   /* real  user  identification */
        int     euid;                   /* eff.  user  identification */
        int     suid;                   /* saved user  identification */
        int     fsuid;                  /* fs    user  identification */
        int     rgid;                   /* real  group identification */
        int     egid;                   /* eff.  group identification */
        int     sgid;                   /* saved group identification */
        int     fsgid;                  /* fs    group identification */
        int     nthr;                   /* number of threads in tgroup */
        char    name[PNAMLEN+1];        /* process name string */
        char    isproc;                 /* boolean: process level? */
        char    state;                  /* process state ('E' = exited) */
        int     excode;                 /* process exit status */
        time_t  btime;                  /* process start time (epoch) */
        time_t  elaps;                  /* process elaps time (hertz) */
        char    cmdline[CMDLEN+1];      /* command-line string */
        int     nthrslpi;               /* # threads in state 'S' */
        int     nthrslpu;               /* # threads in state 'D' */
        int     nthrrun;                /* # threads in state 'R' */
        int     ctid;                   /* OpenVZ container ID */
        int     vpid;                   /* OpenVZ virtual PID */
        int     wasinactive;            /* boolean: task inactive */
        char    container[16];          /* Docker container id (12 pos) */
    } gen;

    /* CPU STATISTICS */
    struct cpu {
        count_t  utime;                 /* time user   text (ticks) */
        count_t  stime;                 /* time system text (ticks) */
        int      nice;                  /* nice value */
        int      prio;                  /* priority */
        int      rtprio;                /* realtime priority */
        int      policy;                /* scheduling policy */
        int      curcpu;                /* current processor */
        int      sleepavg;              /* sleep average percentage */
        int      ifuture[4];            /* reserved for future use */
        char     wchan[16];             /* wait channel string */
        count_t  rundelay;              /* schedstat rundelay (nanosec) */
        count_t  cfuture[1];            /* reserved for future use */
    } cpu;

    /* DISK STATISTICS */
    struct dsk {
        count_t  rio;                   /* number of read requests */
        count_t  rsz;                   /* cumulative # sectors read */
        count_t  wio;                   /* number of write requests */
        count_t  wsz;                   /* cumulative # sectors written */
        count_t  cwsz;                  /* cumulative # written sectors */
        count_t  cfuture[4];            /* reserved for future use */
    } dsk;

    /* MEMORY STATISTICS */
    struct mem {
        count_t  minflt;                /* number of page-reclaims */
        count_t  majflt;                /* number of page-faults */
        count_t  vexec;                 /* virtmem execfile (Kb) */
        count_t  vmem;                  /* virtual  memory  (Kb) */
        count_t  rmem;                  /* resident memory  (Kb) */
        count_t  pmem;                  /* resident memory  (Kb) */
        count_t  vgrow;                 /* virtual  growth  (Kb) */
        count_t  rgrow;                 /* resident growth  (Kb) */
        count_t  vdata;                 /* virtmem data     (Kb) */
        count_t  vstack;                /* virtmem stack    (Kb) */
        count_t  vlibs;                 /* virtmem libexec  (Kb) */
        count_t  vswap;                 /* swap space used  (Kb) */
        count_t  vlock;                 /* virtual locked   (Kb) */
        count_t  cfuture[3];            /* reserved for future use */
    } mem;

    /* NETWORK STATISTICS */
    struct net {
        count_t  tcpsnd;                /* number of TCP-packets sent */
        count_t  tcpssz;                /* cumulative size packets sent */
        count_t  tcprcv;                /* number of TCP-packets recved */
        count_t  tcprsz;                /* cumulative size packets rcvd */
        count_t  udpsnd;                /* number of UDP-packets sent */
        count_t  udpssz;                /* cumulative size packets sent */
        count_t  udprcv;                /* number of UDP-packets recved */
        count_t  udprsz;                /* cumulative size packets sent */
        count_t  avail1;
        count_t  avail2;
        count_t  cfuture[4];            /* reserved for future use */
    } net;

    struct gpu {
        char     state;                 /* A - active, E - Exit, '\0' - no use */
        char     cfuture[3];
        short    nrgpus;                /* number of GPUs for this process */
        int32_t  gpulist;               /* bitlist with GPU numbers */
        int      gpubusy;               /* gpu busy perc process lifetime      -1 = n/a */
        int      membusy;               /* memory busy perc process lifetime   -1 = n/a */
        count_t  timems;                /* milliseconds accounting -1 = n/a, value 0 for active process, value > 0 after termination */
        count_t  memnow;                /* current    memory consumption in KiB */
        count_t  memcum;                /* cumulative memory consumption in KiB */
        count_t  sample;                /* number of samples */
    } gpu;
};
"""  # noqa: E501

c_atop = cstruct().load(atop_def)
c_atop.load(atop_tstat_def, align=True)

AtopRecord = TargetRecordDescriptor(
    "linux/log/atop",
    [
        ("datetime", "ts"),
        ("string", "process"),
        ("string", "cmdline"),
        ("varint", "tgid"),
        ("varint", "pid"),
        ("varint", "ppid"),
        ("varint", "ruid"),
        ("varint", "euid"),
        ("varint", "suid"),
        ("varint", "fsuid"),
        ("varint", "rgid"),
        ("varint", "egid"),
        ("varint", "sgid"),
        ("varint", "fsgid"),
        ("varint", "nthr"),
        ("boolean", "isproc"),
        ("string", "state"),
        ("varint", "excode"),
        ("varint", "elaps"),
        ("varint", "nthrslpi"),
        ("varint", "nthrslpu"),
        ("varint", "nthrrun"),
        ("varint", "ctid"),
        ("varint", "vpid"),
        ("boolean", "wasinactive"),
        ("string", "container"),
        ("path", "filepath"),
    ],
)


class AtopFile:
    """Parse general task information of processes of an Atop log file."""

    def __init__(self, fh: BinaryIO):
        fh.seek(0)

        self.fh = fh
        self.header = c_atop.rawheader(self.fh)
        self.version = self.version()

    def __iter__(self) -> Iterator[c_atop.tstat]:
        while True:
            try:
                record = c_atop.rawrecord(self.fh)
                # system level statistics is not parsed
                self.fh.read(record.scomplen)
                process = self.decompress(self.fh.read(record.pcomplen))
                for _ in range(record.ndeviat):
                    yield c_atop.tstat(process)
            except EOFError:  # noqa: PERF203
                break

    def decompress(self, data: bytes) -> bytes:
        return BytesIO(zlib.decompress(data))

    def version(self) -> str:
        major_version = (self.header.aversion >> 8) & 0x7F
        minor_version = self.header.aversion & 0xFF
        return f"{major_version}.{minor_version}"


class AtopPlugin(Plugin):
    """Unix atop plugin."""

    ATOP_GLOB = "atop_*"
    ATOP_MAGIC = 0xFEEDBEEF
    ATOP_PATH = "/var/log/atop"
    ATOP_VERSIONS = ("2.6", "2.7")

    def check_compatible(self) -> None:
        if not self.target.fs.path(self.ATOP_PATH).exists():
            raise UnsupportedPluginError("No ATOP files found")

    @export(record=AtopRecord)
    def atop(self) -> Iterator[AtopRecord]:
        """Return the content of Atop log files.

        An Atop log file contains the activity of all processes that were running during the interval.
        This includes system-level activity related to the CPU, memory, swap, disks and network layers,
        and for every process (and thread) it shows e.g. the CPU utilization, memory growth, disk utilization,
        priority, username, state, and exit code.

        References:
            - https://diablohorn.com/2022/11/17/parsing-atop-files-with-python-dissect-cstruct/

        Yields AtopRecord with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            process (string): The process name.
            cmdline (string): The command-line of the process.
            tgid (varint): The threadgroup of the process
            pid (varint): The proccess identifier of the process.
            ppid (varint): The proccess identifier of the parent-process.
            ruid (varint): The ruid of the process.
            euid (varint): The euid of the process.
            suid (varint): The suid of the process.
            fsuid (varint): The fsuid of the process.
            rgid (varint): The rgid of the process.
            egid (varint): The egid of the process.
            sgid (varint): The sgid of the process.
            fsgid (varint): The fsgid of the process.
            nthr (varint): The nthr of the process.
            isproc (boolean). The process-level of the process.
            state (string). The state of the process.
            excode (varint): The exit-code of the process.
            elaps (varint): The elapsed time of the process.
            nthrslpi (varint): The threads in state 'S' of the process.
            nthrslpu (varint): The threads in state 'D' of the process.
            nthrrun (varint): The threads in state 'R' of the process.
            ctid (varint): The OpenVZ container ID of the process.
            vpid (varint): The OpenVZ virtual pid of the process.
            wasinactive (boolean): The activity of the process.
            container (string): The Docker Container ID of the process.
            filepath (path): The file name.
        """
        for file in self.target.fs.path(self.ATOP_PATH).glob(self.ATOP_GLOB):
            fh = file.open()

            atop_magic = int.from_bytes(fh.read(4), "little")

            if atop_magic != self.ATOP_MAGIC:
                self.target.log.warning("The Atop log file %s has an invalid magic header", file.name)
                continue

            atop = AtopFile(fh)

            if atop.version not in self.ATOP_VERSIONS:
                self.target.log.warning(
                    "The version %s of the Atop log file %s is incompatible",
                    atop.version,
                    file.name,
                )
                continue

            for entry in atop:
                yield AtopRecord(
                    ts=entry.gen.btime,
                    process=entry.gen.name.decode().strip("\x00"),
                    cmdline=entry.gen.cmdline.decode().strip("\x00"),
                    tgid=entry.gen.tgid,
                    pid=entry.gen.pid,
                    ppid=entry.gen.ppid,
                    ruid=entry.gen.ruid,
                    euid=entry.gen.euid,
                    suid=entry.gen.suid,
                    fsuid=entry.gen.fsuid,
                    rgid=entry.gen.rgid,
                    egid=entry.gen.egid,
                    sgid=entry.gen.sgid,
                    fsgid=entry.gen.fsgid,
                    nthr=entry.gen.nthr,
                    isproc=int.from_bytes(entry.gen.isproc, "little"),
                    state=entry.gen.state.decode().strip("\x00"),
                    excode=entry.gen.excode,
                    elaps=entry.gen.elaps,
                    nthrslpi=entry.gen.nthrslpi,
                    nthrslpu=entry.gen.nthrslpu,
                    nthrrun=entry.gen.nthrrun,
                    ctid=entry.gen.ctid,
                    vpid=entry.gen.vpid,
                    wasinactive=entry.gen.wasinactive,
                    container=entry.gen.container.decode().strip("\x00"),
                    filepath=file,
                    _target=self.target,
                )



from __future__ import annotations

import ipaddress
import struct
from typing import TYPE_CHECKING, NamedTuple

from dissect.cstruct import cstruct
from dissect.util.ts import from_unix

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, alias, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime
    from pathlib import Path

    from dissect.target.target import Target

UTMP_FIELDS = [
    ("datetime", "ts"),
    ("string", "ut_type"),
    ("string", "ut_user"),
    ("varint", "ut_pid"),
    ("string", "ut_line"),
    ("string", "ut_id"),
    ("string", "ut_host"),
    ("net.ipaddress", "ut_addr"),
]

BtmpRecord = TargetRecordDescriptor(
    "linux/log/btmp",
    UTMP_FIELDS,
)

WtmpRecord = TargetRecordDescriptor(
    "linux/log/wtmp",
    UTMP_FIELDS,
)

utmp_def = """
#define UT_LINESIZE     32
#define UT_NAMESIZE     32
#define UT_HOSTSIZE     256

typedef uint32 pid_t;

enum Type : uint8_t {
    EMPTY           = 0x0,
    RUN_LVL         = 0x1,
    BOOT_TIME       = 0x2,
    NEW_TIME        = 0x3,
    OLD_TIME        = 0x4,
    INIT_PROCESS    = 0x5,
    LOGIN_PROCESS   = 0x6,
    USER_PROCESS    = 0x7,
    DEAD_PROCESS    = 0x8,
    ACCOUNTING      = 0x9,
};

struct exit_status {
    uint16 e_termination;
    uint16 e_exit;
};

struct {
    uint32 tv_sec;
    uint32 tv_usec;
} timeval;

struct entry {
    uint32  ut_type;
    pid_t   ut_pid;
    char    ut_line[UT_LINESIZE];
    char    ut_id[4];
    char    ut_user[UT_NAMESIZE];
    char    ut_host[UT_HOSTSIZE];
    struct  exit_status ut_exit;
    long    ut_session;
    struct  timeval ut_tv;
    int32_t ut_addr_v6[4];         // Internet address of remote host; IPv4 address uses just ut_addr_v6[0]
    char    __unused[20];
};
"""

c_utmp = cstruct().load(utmp_def)


class UtmpRecord(NamedTuple):
    ts: datetime
    ut_type: str
    ut_user: str
    ut_pid: int
    ut_line: str
    ut_id: str
    ut_host: str
    ut_addr: ipaddress.IPv4Address | ipaddress.IPv6Address | None


class UtmpFile:
    """Parser for utmp files."""

    def __init__(self, path: Path):
        self.fh = open_decompress(path, "rb")

    def __iter__(self) -> Iterator[UtmpRecord]:
        while True:
            try:
                entry = c_utmp.entry(self.fh)

                r_type = ""
                if entry.ut_type in c_utmp.Type:
                    r_type = c_utmp.Type(entry.ut_type).name

                ut_host = entry.ut_host.decode(errors="surrogateescape").strip("\x00")
                ut_addr = None

                # UTMP misuses the field ut_addr_v6 for IPv4 and IPv6 addresses, because of this the ut_host field
                # is used to determine if the ut_addr_v6 is an IPv6 address where the last 12 bytes of trailing zeroes.
                if entry.ut_addr_v6:
                    if entry.ut_addr_v6[1:] != [0, 0, 0]:
                        # IPv6 address that uses > 4 bytes
                        ut_addr = ipaddress.ip_address(struct.pack("<4i", *entry.ut_addr_v6))
                    else:
                        try:
                            if isinstance(ipaddress.ip_address(ut_host), ipaddress.IPv6Address):
                                # IPv6 address that uses 4 bytes with 12 bytes of trailing zeroes.
                                ut_addr = ipaddress.ip_address(struct.pack("<4i", *entry.ut_addr_v6))
                            elif isinstance(ipaddress.ip_address(ut_host), ipaddress.IPv4Address):
                                # IPv4 address (ut_host, ut_addr_v6)
                                ut_addr = ipaddress.ip_address(struct.pack("<i", entry.ut_addr_v6[0]))
                            else:
                                pass
                        except ValueError:
                            # NOTE: in case the ut_host does not contain a valid IPv6 address,
                            # ut_addr_v6 is parsed as IPv4 address. This could not lead to incorrect results.
                            ut_addr = ipaddress.ip_address(struct.pack("<i", entry.ut_addr_v6[0]))

                yield UtmpRecord(
                    ts=from_unix(entry.ut_tv.tv_sec),
                    ut_type=r_type,
                    ut_pid=entry.ut_pid,
                    ut_user=entry.ut_user.decode(errors="surrogateescape").strip("\x00"),
                    ut_line=entry.ut_line.decode(errors="surrogateescape").strip("\x00"),
                    ut_id=entry.ut_id.decode(errors="surrogateescape").strip("\x00"),
                    ut_host=ut_host,
                    ut_addr=ut_addr,
                )

            except EOFError:  # noqa: PERF203
                break


class UtmpPlugin(Plugin):
    """Unix utmp log plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.btmp_paths = list(self.target.fs.path("/").glob("var/log/btmp*"))
        self.wtmp_paths = list(self.target.fs.path("/").glob("var/log/wtmp*"))
        self.utmp_paths = list(self.target.fs.path("/").glob("var/run/utmp*"))

    def check_compatible(self) -> None:
        if not any(self.btmp_paths + self.wtmp_paths + self.utmp_paths):
            raise UnsupportedPluginError("No wtmp and/or btmp log files found")

    def _build_record(self, record: TargetRecordDescriptor, entry: UtmpRecord) -> Iterator[BtmpRecord | WtmpRecord]:
        return record(
            ts=entry.ts,
            ut_type=entry.ut_type,
            ut_pid=entry.ut_pid,
            ut_user=entry.ut_user,
            ut_line=entry.ut_line,
            ut_id=entry.ut_id,
            ut_host=entry.ut_host,
            ut_addr=entry.ut_addr,
            _target=self.target,
        )

    @export(record=BtmpRecord)
    def btmp(self) -> Iterator[BtmpRecord]:
        """Return failed login attempts stored in the btmp file.

        On a Linux system, failed login attempts are stored in the btmp file located in the var/log/ folder.

        References:
            - https://en.wikipedia.org/wiki/Utmp
            - https://www.thegeekdiary.com/what-is-the-purpose-of-utmp-wtmp-and-btmp-files-in-linux/
        """
        for path in self.btmp_paths:
            if not path.is_file():
                self.target.log.warning("Unable to parse btmp file: %s is not a file", path)
                continue

            for entry in UtmpFile(path):
                yield self._build_record(BtmpRecord, entry)

    @alias("utmp")
    @export(record=WtmpRecord)
    def wtmp(self) -> Iterator[WtmpRecord]:
        """Yield contents of wtmp log files.

        The wtmp file contains the historical data of the utmp file. The utmp file contains information about users
        logins at which terminals, logouts, system events and current status of the system, system boot time
        (used by uptime) etc.

        References:
            - https://www.thegeekdiary.com/what-is-the-purpose-of-utmp-wtmp-and-btmp-files-in-linux/
        """

        for path in self.wtmp_paths + self.utmp_paths:
            if not path.is_file():
                self.target.log.warning("Unable to parse wtmp file: %s is not a file", path)
                continue

            for entry in UtmpFile(path):
                yield self._build_record(WtmpRecord, entry)



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.locate.locate import BaseLocatePlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

gnulocate_def = """
#define MAGIC 0x004c4f43415445303200             /* b'/x00LOCATE02/x00' */

struct entry {
    int8 offset;
    char path[];
}
"""

GNULocateRecord = TargetRecordDescriptor(
    "linux/locate/gnulocate",
    [
        ("path", "path"),
        ("string", "source"),
    ],
)

c_gnulocate = cstruct().load(gnulocate_def)


class GNULocateFile:
    """GNU locate file parser.

    Multiple formats exist for the locatedb file. This class only supports the most recent version ``LOCATE02``.

    The file is encoded with front compression (incremental encoding). This is a form of compression
    which takes a number of characters of the previous encoded entries. Entries are separated with a null byte.

    References:
        - https://manpages.ubuntu.com/manpages/trusty/en/man5/locatedb.5.html
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.count = 0
        self.previous_path = ""

        magic = int.from_bytes(self.fh.read(10), byteorder="big")
        if magic != c_gnulocate.MAGIC:
            raise ValueError(f"Invalid Locate file magic. Expected /x00LOCATE02/x00, got {magic}")

    def __iter__(self) -> Iterator[GNULocateFile]:
        try:
            while True:
                # NOTE: The offset could be negative, which indicates
                # that we decrease the number of characters of the previous path.
                entry = c_gnulocate.entry(self.fh)
                current_filepath_end = entry.path.decode(errors="backslashreplace")
                offset = entry.offset

                self.count += offset

                path = self.previous_path[0 : self.count] + current_filepath_end
                self.previous_path = path
                yield path
        except EOFError:
            return


class GNULocatePlugin(BaseLocatePlugin):
    """GNU locate plugin."""

    __namespace__ = "gnulocate"

    path = "/var/cache/locate/locatedb"

    def check_compatible(self) -> None:
        if not self.target.fs.path(self.path).exists():
            raise UnsupportedPluginError(f"No locatedb file found at {self.path}")

    @export(record=GNULocateRecord)
    def locate(self) -> Iterator[GNULocateRecord]:
        """Yield file and directory names from GNU findutils' locatedb file.

        References:
            - https://manpages.debian.org/testing/locate/locatedb.5.en.html
        """
        locate_fh = self.target.fs.path(self.path).open()
        locate_file = GNULocateFile(locate_fh)

        for path in locate_file:
            yield GNULocateRecord(path=self.target.fs.path(path), source=self.path)



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.filesystems.config import (
    ConfigurationEntry,
    ConfigurationFilesystem,
)
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.plugin import Plugin, internal

if TYPE_CHECKING:
    from collections.abc import Iterable

    from dissect.target.target import Target


class ConfigurationTreePlugin(Plugin):
    __namespace__ = "config_tree"

    def __init__(self, target: Target, dir_path: str = "/"):
        super().__init__(target)
        self.dir_path = dir_path
        self.config_fs = None

        target_dir_path = self.target.fs.path(dir_path)
        if target_dir_path.is_dir():
            self.config_fs = ConfigurationFilesystem(target, dir_path)

        self._get = lru_cache(128)(self._get)

    def check_compatible(self) -> None:
        # This should be able to be retrieved, regardless of OS
        if self.config_fs is None:
            raise UnsupportedPluginError(f"{self.dir_path!r} could not be found.")
        return

    def __call__(
        self,
        path: TargetPath | str | None = None,
        hint: str | None = None,
        collapse: bool | Iterable[str] | None = None,
        collapse_inverse: bool | None = None,
        separator: tuple[str] | None = None,
        comment_prefixes: tuple[str] | None = None,
        as_dict: bool = False,
    ) -> ConfigurationFilesystem | ConfigurationEntry | dict:
        """Create a configuration entry from a file, or a :class:`.ConfigurationFilesystem` from a directory.

        If a directory is specified in ``path``, the other arguments should be provided in the ``get`` call if needed.

        Args:
            path: The path to either a directory or file.
            hint: What kind of parser it should use.
            collapse: Whether it should collapse everything or just a certain set of keys.
            collapse_inverse: Invert the collapse function to collapse everything but the keys inside ``collapse``.
            separator: The separator that should be used for parsing.
            comment_prefixes: What is specified as a comment.
            as_dict: Return a dictionary instead of an entry.
        """
        return self.get(
            path=path,
            as_dict=as_dict,
            hint=hint,
            collapse=collapse,
            collapse_inverse=collapse_inverse,
            separator=separator,
            comment_prefixes=comment_prefixes,
        )

    @internal
    def get(
        self, path: TargetPath | str | None = None, as_dict: bool = False, *args, **kwargs
    ) -> ConfigurationFilesystem | ConfigurationEntry | dict:
        if collapse := kwargs.pop("collapse", None):
            kwargs.update({"collapse": frozenset(collapse)})

        return self._get(path, as_dict, *args, **kwargs)

    def _get(
        self, path: TargetPath | str | None = None, as_dict: bool = False, *args, **kwargs
    ) -> ConfigurationFilesystem | ConfigurationEntry | dict:
        if not path:
            return self.config_fs

        if isinstance(path, TargetPath):
            path = str(path)

        entry = self.config_fs.get(path, None, *args, **kwargs)

        if as_dict:
            return entry.as_dict()

        return entry

    get.__doc__ = __call__.__doc__
